# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# Copyright (C) Gabriel Potter <gabriel@potter.fr>
# This program is published under a GPLv2 license

# flake8: noqa E266
# (We keep comment boxes, it's then one-line comments)

"""
C API calls to Windows DLLs
"""

from pocsuite3.thirdparty.scapy.consts import LINUX, SOLARIS, WINDOWS, BSD,WINDOWS_XP
import ctypes

class bpf_insn(ctypes.Structure):
    """"The BPF instruction data structure"""
    _fields_ = [("code", ctypes.c_ushort),
                ("jt", ctypes.c_ubyte),
                ("jf", ctypes.c_ubyte),
                ("k", ctypes.c_int)]


class bpf_program(ctypes.Structure):
    """"Structure for BIOCSETF"""
    _fields_ = [('bf_len', ctypes.c_int),
                ('bf_insns', ctypes.POINTER(bpf_insn))]


if WINDOWS:
    import ctypes
    import ctypes.wintypes
    from ctypes import Structure, POINTER, byref, create_string_buffer, WINFUNCTYPE
    from pocsuite3.thirdparty.scapy.config import conf

    class bpf_insn(ctypes.Structure):
        """"The BPF instruction data structure"""
        _fields_ = [("code", ctypes.c_ushort),
                    ("jt", ctypes.c_ubyte),
                    ("jf", ctypes.c_ubyte),
                    ("k", ctypes.c_int)]


    class bpf_program(ctypes.Structure):
        """"Structure for BIOCSETF"""
        _fields_ = [('bf_len', ctypes.c_int),
                    ('bf_insns', ctypes.POINTER(bpf_insn))]


    ANY_SIZE = 65500  # FIXME quite inefficient :/
    AF_UNSPEC = 0
    NO_ERROR = 0x0

    CHAR = ctypes.c_char
    DWORD = ctypes.wintypes.DWORD
    BOOL = ctypes.wintypes.BOOL
    BOOLEAN = ctypes.wintypes.BOOLEAN
    ULONG = ctypes.wintypes.ULONG
    ULONGLONG = ctypes.c_ulonglong
    HANDLE = ctypes.wintypes.HANDLE
    LPWSTR = ctypes.wintypes.LPWSTR
    VOID = ctypes.c_void_p
    INT = ctypes.c_int
    UINT = ctypes.wintypes.UINT
    UINT8 = ctypes.c_uint8
    UINT16 = ctypes.c_uint16
    UINT32 = ctypes.c_uint32
    UINT64 = ctypes.c_uint64
    BYTE = ctypes.c_byte
    UCHAR = UBYTE = ctypes.c_ubyte
    SHORT = ctypes.c_short
    USHORT = ctypes.c_ushort


    # UTILS


    def _resolve_list(list_obj):
        current = list_obj
        _list = []
        while current and hasattr(current, "contents"):
            _list.append(_struct_to_dict(current.contents))
            current = current.contents.next
        return _list


    def _struct_to_dict(struct_obj):
        results = {}
        for fname, ctype in struct_obj.__class__._fields_:
            val = getattr(struct_obj, fname)
            if fname == "next":
                # Already covered by the trick below
                continue
            if issubclass(ctype, (Structure, ctypes.Union)):
                results[fname] = _struct_to_dict(val)
            elif val and hasattr(val, "contents"):
                # Let's resolve recursive pointers
                if hasattr(val.contents, "next"):
                    results[fname] = _resolve_list(val)
                else:
                    results[fname] = val
            else:
                results[fname] = val
        return results

    ##############################
    ####### WinAPI handles #######
    ##############################

    _winapi_SetConsoleTitle = ctypes.windll.kernel32.SetConsoleTitleW
    _winapi_SetConsoleTitle.restype = BOOL
    _winapi_SetConsoleTitle.argtypes = [LPWSTR]

    def _windows_title(title=None):
        """Updates the terminal title with the default one or with `title`
        if provided."""
        if conf.interactive:
            _winapi_SetConsoleTitle(title or "Scapy v{}".format(conf.version))


    SC_HANDLE = HANDLE

    class SERVICE_STATUS(Structure):
        """https://docs.microsoft.com/en-us/windows/desktop/api/winsvc/ns-winsvc-_service_status"""  # noqa: E501
        _fields_ = [("dwServiceType", DWORD),
                    ("dwCurrentState", DWORD),
                    ("dwControlsAccepted", DWORD),
                    ("dwWin32ExitCode", DWORD),
                    ("dwServiceSpecificExitCode", DWORD),
                    ("dwCheckPoint", DWORD),
                    ("dwWaitHint", DWORD)]


    OpenServiceW = ctypes.windll.Advapi32.OpenServiceW
    OpenServiceW.restype = SC_HANDLE
    OpenServiceW.argtypes = [SC_HANDLE, LPWSTR, DWORD]

    CloseServiceHandle = ctypes.windll.Advapi32.CloseServiceHandle
    CloseServiceHandle.restype = BOOL
    CloseServiceHandle.argtypes = [SC_HANDLE]

    OpenSCManagerW = ctypes.windll.Advapi32.OpenSCManagerW
    OpenSCManagerW.restype = SC_HANDLE
    OpenSCManagerW.argtypes = [LPWSTR, LPWSTR, DWORD]

    QueryServiceStatus = ctypes.windll.Advapi32.QueryServiceStatus
    QueryServiceStatus.restype = BOOL
    QueryServiceStatus.argtypes = [SC_HANDLE, POINTER(SERVICE_STATUS)]

    def get_service_status(service):
        """Returns content of QueryServiceStatus for a service"""
        SERVICE_QUERY_STATUS = 0x0004
        schSCManager = OpenSCManagerW(
            None,  # Local machine
            None,  # SERVICES_ACTIVE_DATABASE
            SERVICE_QUERY_STATUS
        )
        service = OpenServiceW(
            schSCManager,
            service,
            SERVICE_QUERY_STATUS
        )
        status = SERVICE_STATUS()
        QueryServiceStatus(
            service,
            status
        )
        result = _struct_to_dict(status)
        CloseServiceHandle(service)
        CloseServiceHandle(schSCManager)
        return result


    ##############################
    ###### Define IPHLPAPI  ######
    ##############################


    iphlpapi = ctypes.windll.iphlpapi

    ##############################
    ########### Common ###########
    ##############################


    class in_addr(Structure):
        _fields_ = [("byte", UBYTE * 4)]


    class in6_addr(Structure):
        _fields_ = [("byte", UBYTE * 16)]


    class sockaddr_in(Structure):
        _fields_ = [("sin_family", SHORT),
                    ("sin_port", USHORT),
                    ("sin_addr", in_addr),
                    ("sin_zero", 8 * CHAR)]


    class sockaddr_in6(Structure):
        _fields_ = [("sin6_family", SHORT),
                    ("sin6_port", USHORT),
                    ("sin6_flowinfo", ULONG),
                    ("sin6_addr", in6_addr),
                    ("sin6_scope_id", ULONG)]


    class SOCKADDR_INET(ctypes.Union):
        _fields_ = [("Ipv4", sockaddr_in),
                    ("Ipv6", sockaddr_in6),
                    ("si_family", USHORT)]

    ##############################
    ######### ICMP stats #########
    ##############################


    class MIBICMPSTATS(Structure):
        _fields_ = [("dwMsgs", DWORD),
                    ("dwErrors", DWORD),
                    ("dwDestUnreachs", DWORD),
                    ("dwTimeExcds", DWORD),
                    ("dwParmProbs", DWORD),
                    ("dwSrcQuenchs", DWORD),
                    ("dwRedirects", DWORD),
                    ("dwEchos", DWORD),
                    ("dwEchoReps", DWORD),
                    ("dwTimestamps", DWORD),
                    ("dwTimestampReps", DWORD),
                    ("dwAddrMasks", DWORD),
                    ("dwAddrMaskReps", DWORD)]


    class MIBICMPINFO(Structure):
        _fields_ = [("icmpInStats", MIBICMPSTATS),
                    ("icmpOutStats", MIBICMPSTATS)]


    class MIB_ICMP(Structure):
        _fields_ = [("stats", MIBICMPINFO)]


    PMIB_ICMP = POINTER(MIB_ICMP)

    # Func

    _GetIcmpStatistics = WINFUNCTYPE(ULONG, PMIB_ICMP)(
        ('GetIcmpStatistics', iphlpapi))


    def GetIcmpStatistics():
        """Return all Windows ICMP stats from iphlpapi"""
        statistics = MIB_ICMP()
        _GetIcmpStatistics(byref(statistics))
        results = _struct_to_dict(statistics)
        del(statistics)
        return results

    ##############################
    ##### Adapters Addresses #####
    ##############################


    # Our GetAdaptersAddresses implementation is inspired by
    # @sphaero 's gist: https://gist.github.com/sphaero/f9da6ebb9a7a6f679157
    # published under a MPL 2.0 License (GPLv2 compatible)

    # from iptypes.h
    MAX_ADAPTER_ADDRESS_LENGTH = 8
    MAX_DHCPV6_DUID_LENGTH = 130

    GAA_FLAG_INCLUDE_PREFIX = ULONG(0x0010)
    # for now, just use void * for pointers to unused structures
    PIP_ADAPTER_WINS_SERVER_ADDRESS_LH = VOID
    PIP_ADAPTER_GATEWAY_ADDRESS_LH = VOID
    PIP_ADAPTER_DNS_SUFFIX = VOID

    IF_OPER_STATUS = UINT
    IF_LUID = UINT64

    NET_IF_COMPARTMENT_ID = UINT32
    GUID = BYTE * 16
    NET_IF_NETWORK_GUID = GUID
    NET_IF_CONNECTION_TYPE = UINT  # enum
    TUNNEL_TYPE = UINT  # enum


    class SOCKET_ADDRESS(ctypes.Structure):
        _fields_ = [('address', POINTER(SOCKADDR_INET)),
                    ('length', INT)]


    class _IP_ADAPTER_ADDRESSES_METRIC(Structure):
        _fields_ = [('length', ULONG),
                    ('interface_index', DWORD)]


    class IP_ADAPTER_UNICAST_ADDRESS(Structure):
        pass


    PIP_ADAPTER_UNICAST_ADDRESS = POINTER(IP_ADAPTER_UNICAST_ADDRESS)
    if WINDOWS_XP:
        IP_ADAPTER_UNICAST_ADDRESS._fields_ = [
            ("length", ULONG),
            ("flags", DWORD),
            ("next", PIP_ADAPTER_UNICAST_ADDRESS),
            ("address", SOCKET_ADDRESS),
            ("prefix_origin", INT),
            ("suffix_origin", INT),
            ("dad_state", INT),
            ("valid_lifetime", ULONG),
            ("preferred_lifetime", ULONG),
            ("lease_lifetime", ULONG),
        ]
    else:
        IP_ADAPTER_UNICAST_ADDRESS._fields_ = [
            ("length", ULONG),
            ("flags", DWORD),
            ("next", PIP_ADAPTER_UNICAST_ADDRESS),
            ("address", SOCKET_ADDRESS),
            ("prefix_origin", INT),
            ("suffix_origin", INT),
            ("dad_state", INT),
            ("valid_lifetime", ULONG),
            ("preferred_lifetime", ULONG),
            ("lease_lifetime", ULONG),
            ("on_link_prefix_length", UBYTE)
        ]


    class IP_ADAPTER_ANYCAST_ADDRESS(Structure):
        pass


    PIP_ADAPTER_ANYCAST_ADDRESS = POINTER(IP_ADAPTER_ANYCAST_ADDRESS)
    IP_ADAPTER_ANYCAST_ADDRESS._fields_ = [
        ("length", ULONG),
        ("flags", DWORD),
        ("next", PIP_ADAPTER_ANYCAST_ADDRESS),
        ("address", SOCKET_ADDRESS),
    ]


    class IP_ADAPTER_MULTICAST_ADDRESS(Structure):
        pass


    PIP_ADAPTER_MULTICAST_ADDRESS = POINTER(IP_ADAPTER_MULTICAST_ADDRESS)
    IP_ADAPTER_MULTICAST_ADDRESS._fields_ = [
        ("length", ULONG),
        ("flags", DWORD),
        ("next", PIP_ADAPTER_MULTICAST_ADDRESS),
        ("address", SOCKET_ADDRESS),
    ]


    class IP_ADAPTER_DNS_SERVER_ADDRESS(Structure):
        pass


    PIP_ADAPTER_DNS_SERVER_ADDRESS = POINTER(IP_ADAPTER_DNS_SERVER_ADDRESS)
    IP_ADAPTER_DNS_SERVER_ADDRESS._fields_ = [
        ("length", ULONG),
        ("flags", DWORD),
        ("next", PIP_ADAPTER_DNS_SERVER_ADDRESS),
        ("address", SOCKET_ADDRESS),
    ]


    class IP_ADAPTER_PREFIX(Structure):
        pass


    PIP_ADAPTER_PREFIX = ctypes.POINTER(IP_ADAPTER_PREFIX)
    IP_ADAPTER_PREFIX._fields_ = [
        ("alignment", ULONGLONG),
        ("next", PIP_ADAPTER_PREFIX),
        ("address", SOCKET_ADDRESS),
        ("prefix_length", ULONG)
    ]


    class IP_ADAPTER_ADDRESSES(Structure):
        pass


    LP_IP_ADAPTER_ADDRESSES = POINTER(IP_ADAPTER_ADDRESSES)

    if WINDOWS_XP:
        IP_ADAPTER_ADDRESSES._fields_ = [
            ('length', ULONG),
            ('interface_index', DWORD),
            ('next', LP_IP_ADAPTER_ADDRESSES),
            ('adapter_name', ctypes.c_char_p),
            ('first_unicast_address', PIP_ADAPTER_UNICAST_ADDRESS),
            ('first_anycast_address', PIP_ADAPTER_ANYCAST_ADDRESS),
            ('first_multicast_address', PIP_ADAPTER_MULTICAST_ADDRESS),
            ('first_dns_server_address', PIP_ADAPTER_DNS_SERVER_ADDRESS),
            ('dns_suffix', ctypes.c_wchar_p),
            ('description', ctypes.c_wchar_p),
            ('friendly_name', ctypes.c_wchar_p),
            ('physical_address', BYTE * MAX_ADAPTER_ADDRESS_LENGTH),
            ('physical_address_length', ULONG),
            ('flags', ULONG),
            ('mtu', ULONG),
            ('interface_type', DWORD),
            ('oper_status', IF_OPER_STATUS),
            ('ipv6_interface_index', DWORD),
            ('zone_indices', ULONG * 16),
            ('first_prefix', PIP_ADAPTER_PREFIX),
        ]
    else:
        IP_ADAPTER_ADDRESSES._fields_ = [
            ('length', ULONG),
            ('interface_index', DWORD),
            ('next', LP_IP_ADAPTER_ADDRESSES),
            ('adapter_name', ctypes.c_char_p),
            ('first_unicast_address', PIP_ADAPTER_UNICAST_ADDRESS),
            ('first_anycast_address', PIP_ADAPTER_ANYCAST_ADDRESS),
            ('first_multicast_address', PIP_ADAPTER_MULTICAST_ADDRESS),
            ('first_dns_server_address', PIP_ADAPTER_DNS_SERVER_ADDRESS),
            ('dns_suffix', ctypes.c_wchar_p),
            ('description', ctypes.c_wchar_p),
            ('friendly_name', ctypes.c_wchar_p),
            ('physical_address', BYTE * MAX_ADAPTER_ADDRESS_LENGTH),
            ('physical_address_length', ULONG),
            ('flags', ULONG),
            ('mtu', ULONG),
            ('interface_type', DWORD),
            ('oper_status', IF_OPER_STATUS),
            ('ipv6_interface_index', DWORD),
            ('zone_indices', ULONG * 16),
            ('first_prefix', PIP_ADAPTER_PREFIX),
            ('transmit_link_speed', ULONGLONG),
            ('receive_link_speed', ULONGLONG),
            ('first_wins_server_address', PIP_ADAPTER_WINS_SERVER_ADDRESS_LH),
            ('first_gateway_address', PIP_ADAPTER_GATEWAY_ADDRESS_LH),
            ('ipv4_metric', ULONG),
            ('ipv6_metric', ULONG),
            ('luid', IF_LUID),
            ('dhcpv4_server', SOCKET_ADDRESS),
            ('compartment_id', NET_IF_COMPARTMENT_ID),
            ('network_guid', NET_IF_NETWORK_GUID),
            ('connection_type', NET_IF_CONNECTION_TYPE),
            ('tunnel_type', TUNNEL_TYPE),
            ('dhcpv6_server', SOCKET_ADDRESS),
            ('dhcpv6_client_duid', BYTE * MAX_DHCPV6_DUID_LENGTH),
            ('dhcpv6_client_duid_length', ULONG),
            ('dhcpv6_iaid', ULONG),
            ('first_dns_suffix', PIP_ADAPTER_DNS_SUFFIX)]

    # Func

    _GetAdaptersAddresses = WINFUNCTYPE(ULONG, ULONG, ULONG,
                                        POINTER(VOID),
                                        LP_IP_ADAPTER_ADDRESSES,
                                        POINTER(ULONG))(
                                            ('GetAdaptersAddresses', iphlpapi))


    def GetAdaptersAddresses(AF=AF_UNSPEC):
        """Return all Windows Adapters addresses from iphlpapi"""
        # We get the size first
        size = ULONG()
        flags = GAA_FLAG_INCLUDE_PREFIX
        res = _GetAdaptersAddresses(AF, flags,
                                    None, None,
                                    byref(size))
        if res != 0x6f:  # BUFFER OVERFLOW -> populate size
            raise RuntimeError("Error getting structure length (%d)" % res)
        # Now let's build our buffer
        pointer_type = POINTER(IP_ADAPTER_ADDRESSES)
        buffer = create_string_buffer(size.value)
        AdapterAddresses = ctypes.cast(buffer, pointer_type)
        # And call GetAdaptersAddresses
        res = _GetAdaptersAddresses(AF, flags,
                                    None, AdapterAddresses,
                                    byref(size))
        if res != NO_ERROR:
            raise RuntimeError("Error retrieving table (%d)" % res)
        results = _resolve_list(AdapterAddresses)
        del(AdapterAddresses)
        return results

    ##############################
    ####### Routing tables #######
    ##############################

    ### V1 ###


    class MIB_IPFORWARDROW(Structure):
        _fields_ = [('ForwardDest', DWORD),
                    ('ForwardMask', DWORD),
                    ('ForwardPolicy', DWORD),
                    ('ForwardNextHop', DWORD),
                    ('ForwardIfIndex', DWORD),
                    ('ForwardType', DWORD),
                    ('ForwardProto', DWORD),
                    ('ForwardAge', DWORD),
                    ('ForwardNextHopAS', DWORD),
                    ('ForwardMetric1', DWORD),
                    ('ForwardMetric2', DWORD),
                    ('ForwardMetric3', DWORD),
                    ('ForwardMetric4', DWORD),
                    ('ForwardMetric5', DWORD)]


    class MIB_IPFORWARDTABLE(Structure):
        _fields_ = [('NumEntries', DWORD),
                    ('Table', MIB_IPFORWARDROW * ANY_SIZE)]


    PMIB_IPFORWARDTABLE = POINTER(MIB_IPFORWARDTABLE)

    # Func

    _GetIpForwardTable = WINFUNCTYPE(DWORD,
                                     PMIB_IPFORWARDTABLE, POINTER(ULONG), BOOL)(
                                         ('GetIpForwardTable', iphlpapi))


    def GetIpForwardTable():
        """Return all Windows routes (IPv4 only) from iphlpapi"""
        # We get the size first
        size = ULONG()
        res = _GetIpForwardTable(None, byref(size), False)
        if res != 0x7a:  # ERROR_INSUFFICIENT_BUFFER -> populate size
            raise RuntimeError("Error getting structure length (%d)" % res)
        # Now let's build our buffer
        pointer_type = PMIB_IPFORWARDTABLE
        buffer = create_string_buffer(size.value)
        pIpForwardTable = ctypes.cast(buffer, pointer_type)
        # And call GetAdaptersAddresses
        res = _GetIpForwardTable(pIpForwardTable, byref(size), True)
        if res != NO_ERROR:
            raise RuntimeError("Error retrieving table (%d)" % res)
        results = []
        for i in range(pIpForwardTable.contents.NumEntries):
            results.append(_struct_to_dict(pIpForwardTable.contents.Table[i]))
        del(pIpForwardTable)
        return results

    ### V2 ###


    NET_IFINDEX = ULONG
    NL_ROUTE_PROTOCOL = INT
    NL_ROUTE_ORIGIN = INT


    class NET_LUID(Structure):
        _fields_ = [("Value", ULONGLONG)]


    class IP_ADDRESS_PREFIX(Structure):
        _fields_ = [("Prefix", SOCKADDR_INET),
                    ("PrefixLength", UINT8)]


    class MIB_IPFORWARD_ROW2(Structure):
        _fields_ = [("InterfaceLuid", NET_LUID),
                    ("InterfaceIndex", NET_IFINDEX),
                    ("DestinationPrefix", IP_ADDRESS_PREFIX),
                    ("NextHop", SOCKADDR_INET),
                    ("SitePrefixLength", UCHAR),
                    ("ValidLifetime", ULONG),
                    ("PreferredLifetime", ULONG),
                    ("Metric", ULONG),
                    ("Protocol", NL_ROUTE_PROTOCOL),
                    ("Loopback", BOOLEAN),
                    ("AutoconfigureAddress", BOOLEAN),
                    ("Publish", BOOLEAN),
                    ("Immortal", BOOLEAN),
                    ("Age", ULONG),
                    ("Origin", NL_ROUTE_ORIGIN)]


    class MIB_IPFORWARD_TABLE2(Structure):
        _fields_ = [("NumEntries", ULONG),
                    ("Table", MIB_IPFORWARD_ROW2 * ANY_SIZE)]


    PMIB_IPFORWARD_TABLE2 = POINTER(MIB_IPFORWARD_TABLE2)

    # Func

    if not WINDOWS_XP:
        # GetIpForwardTable2 does not exist under Windows XP
        _GetIpForwardTable2 = WINFUNCTYPE(
            ULONG, USHORT,
            POINTER(PMIB_IPFORWARD_TABLE2))(
                ('GetIpForwardTable2', iphlpapi)
        )
        _FreeMibTable = WINFUNCTYPE(None, PMIB_IPFORWARD_TABLE2)(
            ('FreeMibTable', iphlpapi)
        )


    def GetIpForwardTable2(AF=AF_UNSPEC):
        """Return all Windows routes (IPv4/IPv6) from iphlpapi"""
        if WINDOWS_XP:
            raise OSError("Not available on Windows XP !")
        table = PMIB_IPFORWARD_TABLE2()
        res = _GetIpForwardTable2(AF, byref(table))
        if res != NO_ERROR:
            raise RuntimeError("Error retrieving table (%d)" % res)
        results = []
        for i in range(table.contents.NumEntries):
            results.append(_struct_to_dict(table.contents.Table[i]))
        _FreeMibTable(table)
        return results



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

"""
Operating system specific functionality.
"""

# from __future__ import absolute_import
import socket

from pocsuite3.thirdparty.scapy.consts import LINUX, SOLARIS, WINDOWS, BSD
from pocsuite3.thirdparty.scapy.config import conf, _set_conf_sockets
from pocsuite3.thirdparty.scapy.data import IPV6_ADDR_GLOBAL
from pocsuite3.thirdparty.scapy.compat import orb
from pocsuite3.thirdparty.scapy.pton_ntop import inet_ntop, inet_pton
conf.use_pcap = True

def str2mac(s):
    return ("%02x:" * 6)[:-1] % tuple(orb(x) for x in s)


if not WINDOWS:
    if not conf.use_pcap:
        from pocsuite3.thirdparty.scapy.core import get_if_raw_addr


def get_if_addr(iff):
    return inet_ntop(socket.AF_INET, get_if_raw_addr(iff))


if LINUX:
    from pocsuite3.thirdparty.scapy.linux import *  # noqa F403
elif BSD:
    from pocsuite3.thirdparty.scapy.unix import read_routes, read_routes6, in6_getifaddr  # noqa: F401, E501
    from pocsuite3.thirdparty.scapy.core import *  # noqa F403
    if not conf.use_pcap:
        # Native
        from pocsuite3.thirdparty.scapy.supersocket import * # noqa F403
        conf.use_bpf = True
elif SOLARIS:
    from pocsuite3.thirdparty.scapy.solaris import *  # noqa F403
elif WINDOWS:
    from pocsuite3.thirdparty.scapy.arch_windows_init import *  # noqa F403
    from pocsuite3.thirdparty.scapy.native import *  # noqa F403

if conf.iface is None:
    conf.iface = conf.loopback_name

_set_conf_sockets()  # Apply config


def get_if_addr6(iff):
    """
    Returns the main global unicast address associated with provided
    interface, in human readable form. If no global address is found,
    None is returned.
    """
    return next((x[0] for x in in6_getifaddr()
                 if x[2] == iff and x[1] == IPV6_ADDR_GLOBAL), None)


def get_if_raw_addr6(iff):
    """
    Returns the main global unicast address associated with provided
    interface, in network format. If no global address is found, None
    is returned.
    """
    ip6 = get_if_addr6(iff)
    if ip6 is not None:
        return inet_pton(socket.AF_INET6, ip6)

    return None



import getpass
import urllib

from configparser import ConfigParser
from pocsuite3.lib.core.data import logger, kb
from pocsuite3.lib.core.data import paths
from pocsuite3.lib.request import requests


class ZoomEye():
    def __init__(self, conf_path=paths.POCSUITE_RC_PATH, username=None, password=None):
        self.headers = None
        self.token = None
        self.resources = None
        self.plan = None
        self.conf_path = conf_path
        self.username = username
        self.password = password

        if self.conf_path:
            self.parser = ConfigParser()
            self.parser.read(self.conf_path)

        if username and password:
            self.check_account()
        else:
            try:
                self.token = self.parser.get("Telnet404", 'Jwt token')
            except Exception:
                pass

    def token_is_available(self):
        if self.token:
            headers = {'Authorization': 'JWT %s' % self.token}
            try:
                resp = requests.get('https://api.zoomeye.org/resources-info', headers=headers)
                if resp and resp.status_code == 200 and "plan" in resp.json():
                    self.headers = headers
                    return True
            except Exception as ex:
                logger.error(str(ex))
        return False

    def new_token(self):
        data = '{{"username": "{}", "password": "{}"}}'.format(self.username, self.password)
        try:
            resp = requests.post('https://api.zoomeye.org/user/login', data=data, )
            if resp.status_code != 401 and "access_token" in resp.json():
                content = resp.json()
                self.token = content['access_token']
                self.headers = {'Authorization': 'JWT %s' % self.token}
                return True
        except Exception as ex:
            logger.error(str(ex))
        return False

    def check_account(self):
        if self.token_is_available():
            return True
        else:
            if self.username and self.password:
                if self.new_token():
                    self.write_conf()
                    return True
            else:
                username = input("Telnet404 email account:")
                password = getpass.getpass("Telnet404 password:")
                self.username = username
                self.password = password
                if self.new_token():
                    self.write_conf()
                    return True
                else:
                    logger.error("The username or password is incorrect. "
                                 "Please enter the correct username and password.")
                    return False
        return False

    def write_conf(self):
        if not self.parser.has_section("Telnet404"):
            self.parser.add_section("Telnet404")
        try:
            self.parser.set("Telnet404", "Jwt token", self.token)
            self.parser.write(open(self.conf_path, "w"))
        except Exception as ex:
            logger.error(str(ex))

    def get_resource_info(self):
        if self.check_account():
            try:
                resp = requests.get('https://api.zoomeye.org/resources-info', headers=self.headers)
                if resp and resp.status_code == 200 and 'plan' in resp.json():
                    content = resp.json()
                    self.plan = content['plan']
                    self.resources = content['resources']['search']
                    return True
            except Exception as ex:
                logger.error(str(ex))
        return False

    def search(self, dork, pages=1, resource='web'):
        search_result = set()
        if kb.comparison:
            kb.comparison.add_dork("Zoomeye", dork)
        try:
            for page in range(1, pages + 1):
                url = "https://api.zoomeye.org/{}/search?query={}&page={}&facet=app,os".format(resource,
                                                                                               urllib.parse.quote(dork),
                                                                                               page)
                resp = requests.get(url, headers=self.headers)
                if resp and resp.status_code == 200 and "matches" in resp.json():
                    content = resp.json()
                    if resource == 'web':
                        for match in content["matches"]:
                            ans = match["site"]
                            search_result.add(ans)
                            if kb.comparison:
                                honeypot = False
                                if "honeypot" in content or "honeypot_lastupdate" in content:
                                    honeypot = True
                                kb.comparison.add_ip(ans, "Zoomeye", honeypot)
                    else:
                        for match in content['matches']:
                            ans = match['ip']
                            if 'portinfo' in match:
                                ans += ':' + str(match['portinfo']['port'])
                            search_result.add(ans)
                            if kb.comparison:
                                honeypot = False
                                if "honeypot" in match or "honeypot_lastupdate" in match:
                                    honeypot = True
                                kb.comparison.add_ip(ans, "Zoomeye", honeypot)
        except Exception as ex:
            logger.error(str(ex))
        return search_result


if __name__ == "__main__":
    ze = ZoomEye()
    ze.search('dedecms')



import os
import re
import subprocess


def stdout_encode(data):
    """
    Cross-linked function
    """
    if isinstance(data, bytes):
        data = data.decode('utf-8')
    else:
        data = str(data)
    return data


def get_revision_number():
    """
    Returns abbreviated commit hash number as retrieved with "git rev-parse --short HEAD"
    """

    ret = None
    file_path = None
    _ = os.path.dirname(__file__)

    while True:
        file_path = os.path.join(_, ".git", "HEAD")
        if os.path.exists(file_path):
            break
        else:
            file_path = None
            if _ == os.path.dirname(_):
                break
            else:
                _ = os.path.dirname(_)

    while True:
        if file_path and os.path.isfile(file_path):
            with open(file_path, "r") as f:
                content = f.read()
                file_path = None
                if content.startswith("ref: "):
                    file_path = os.path.join(_, ".git", content.replace("ref: ", "")).strip()
                else:
                    match = re.match(r"(?i)[0-9a-f]{32}", content)
                    ret = match.group(0) if match else None
                    break
        else:
            break

    if not ret:
        process = subprocess.Popen("git rev-parse --verify HEAD",
                                   shell=True,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        stdout, _ = process.communicate()
        stdout = stdout_encode(stdout)
        match = re.search(r"(?i)[0-9a-f]{32}", stdout or "")
        ret = match.group(0) if match else None

    return ret[:7] if ret else None



import sys
import time
import os
import re
import subprocess
from pocsuite3.lib.core.common import data_to_stdout
from pocsuite3.lib.core.common import get_latest_revision
from pocsuite3.lib.core.common import poll_process
from pocsuite3.lib.core.data import conf
from pocsuite3.lib.core.data import logger
from pocsuite3.lib.core.data import paths
from pocsuite3.lib.core.revision import get_revision_number
from pocsuite3.lib.core.settings import GIT_REPOSITORY
from pocsuite3.lib.core.settings import IS_WIN
from pocsuite3.lib.core.settings import UNICODE_ENCODING
from pocsuite3.lib.core.settings import VERSION


def update():
    if not conf.update_all:
        return
    success = False
    if not os.path.exists(os.path.join(paths.POCSUITE_ROOT_PATH, "../", ".git")):
        warn_msg = "not a git repository. It is recommended to clone the 'knownsec/pocsuite3' repository "
        warn_msg += "from GitHub (e.g. 'git clone --depth 1 {} pocsuite3')".format(GIT_REPOSITORY)
        logger.warn(warn_msg)
        if VERSION == get_latest_revision():
            logger.info("already at the latest revision '{}'".format(get_revision_number()))
            return
    else:
        info_msg = "updating pocsuite3 to the latest development revision from the "
        info_msg += "GitHub repository"
        logger.info(info_msg)
        debug_msg = "pocsuite3 will try to update itself using 'git' command"
        logger.debug(debug_msg)
        data_to_stdout("\r[{0}] [INFO] update in progress ".format(time.strftime("%X")))
        cwd_path = os.path.join(paths.POCSUITE_ROOT_PATH, "../")
        try:
            process = subprocess.Popen("git checkout . && git pull %s HEAD" % GIT_REPOSITORY,
                                       shell=True,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       cwd=cwd_path.encode(
                                           sys.getfilesystemencoding() or UNICODE_ENCODING))
            poll_process(process, True)
            stdout, stderr = process.communicate()
            success = not process.returncode
        except (IOError, OSError) as ex:
            success = False
            stderr = str(ex)
        if success:
            logger.info("{0} the latest revision '{1}'".format("already at" if b"Already" in stdout else "updated to",
                                                               get_revision_number()))
        else:
            if "Not a git repository" in stderr:
                err_msg = "not a valid git repository. Please checkout the 'knownsec/pocsuite3' repository "
                err_msg += "from GitHub (e.g. 'git clone --depth 1 %s pocsuite3')" % GIT_REPOSITORY
                logger.error(err_msg)
            else:
                logger.error("update could not be completed ('%s')" % re.sub(r"\W+", " ", stderr).strip())
    if not success:
        if IS_WIN:
            info_msg = "for Windows platform it's recommended "
            info_msg += "to use a GitHub for Windows client for updating "
            info_msg += "purposes (http://windows.github.com/) or just "
            info_msg += "download the latest snapshot from "
            info_msg += "https://github.com/knownsec/pocsuite3/downloads"
        else:
            info_msg = "for Linux platform it's recommended "
            info_msg += "to install a standard 'git' package (e.g.: 'sudo apt-get install git')"
        logger.info(info_msg)
    sys.exit()



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

"""
Convert IPv6 addresses between textual representation and binary.

These functions are missing when python is compiled
without IPv6 support, on Windows for instance.
"""

# from __future__ import absolute_import
import socket
import re
import binascii
from pocsuite3.thirdparty.scapy.six.moves import range
from pocsuite3.thirdparty.scapy.compat import plain_str, hex_bytes, bytes_encode, bytes_hex

_IP6_ZEROS = re.compile('(?::|^)(0(?::0)+)(?::|$)')
_INET6_PTON_EXC = socket.error("illegal IP address string passed to inet_pton")


def _inet6_pton(addr):
    """Convert an IPv6 address from text representation into binary form,
used when socket.inet_pton is not available.

    """
    joker_pos = None
    result = b""
    addr = plain_str(addr)
    if addr == '::':
        return b'\x00' * 16
    if addr.startswith('::'):
        addr = addr[1:]
    if addr.endswith('::'):
        addr = addr[:-1]
    parts = addr.split(":")
    nparts = len(parts)
    for i, part in enumerate(parts):
        if not part:
            # "::" indicates one or more groups of 2 null bytes
            if joker_pos is None:
                joker_pos = len(result)
            else:
                # Wildcard is only allowed once
                raise _INET6_PTON_EXC
        elif i + 1 == nparts and '.' in part:
            # The last part of an IPv6 address can be an IPv4 address
            if part.count('.') != 3:
                # we have to do this since socket.inet_aton('1.2') ==
                # b'\x01\x00\x00\x02'
                raise _INET6_PTON_EXC
            try:
                result += socket.inet_aton(part)
            except socket.error:
                raise _INET6_PTON_EXC
        else:
            # Each part must be 16bit. Add missing zeroes before decoding.
            try:
                result += hex_bytes(part.rjust(4, "0"))
            except (binascii.Error, TypeError):
                raise _INET6_PTON_EXC
    # If there's a wildcard, fill up with zeros to reach 128bit (16 bytes)
    if joker_pos is not None:
        if len(result) == 16:
            raise _INET6_PTON_EXC
        result = (result[:joker_pos] + b"\x00" * (16 - len(result)) +
                  result[joker_pos:])
    if len(result) != 16:
        raise _INET6_PTON_EXC
    return result


_INET_PTON = {
    socket.AF_INET: socket.inet_aton,
    socket.AF_INET6: _inet6_pton,
}


def inet_pton(af, addr):
    """Convert an IP address from text representation into binary form."""
    # Will replace Net/Net6 objects
    addr = plain_str(addr)
    # Use inet_pton if available
    try:
        return socket.inet_pton(af, addr)
    except AttributeError:
        try:
            return _INET_PTON[af](addr)
        except KeyError:
            raise socket.error("Address family not supported by protocol")


def _inet6_ntop(addr):
    """Convert an IPv6 address from binary form into text representation,
used when socket.inet_pton is not available.

    """
    # IPv6 addresses have 128bits (16 bytes)
    if len(addr) != 16:
        raise ValueError("invalid length of packed IP address string")

    # Decode to hex representation
    address = ":".join(plain_str(bytes_hex(addr[idx:idx + 2])).lstrip('0') or '0'  # noqa: E501
                       for idx in range(0, 16, 2))

    try:
        # Get the longest set of zero blocks. We need to take a look
        # at group 1 regarding the length, as 0:0:1:0:0:2:3:4 would
        # have two matches: 0:0: and :0:0: where the latter is longer,
        # though the first one should be taken. Group 1 is in both
        # cases 0:0.
        match = max(_IP6_ZEROS.finditer(address),
                    key=lambda m: m.end(1) - m.start(1))
        return '{}::{}'.format(address[:match.start()], address[match.end():])
    except ValueError:
        return address


_INET_NTOP = {
    socket.AF_INET: socket.inet_ntoa,
    socket.AF_INET6: _inet6_ntop,
}


def inet_ntop(af, addr):
    """Convert an IP address from binary form into text representation."""
    # Use inet_ntop if available
    addr = bytes_encode(addr)
    try:
        return socket.inet_ntop(af, addr)
    except AttributeError:
        try:
            return _INET_NTOP[af](addr)
        except KeyError:
            raise ValueError("unknown address family %d" % af)



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

from sys import platform, maxsize
import platform as platform_lib

LINUX = platform.startswith("linux")
OPENBSD = platform.startswith("openbsd")
FREEBSD = "freebsd" in platform
NETBSD = platform.startswith("netbsd")
DARWIN = platform.startswith("darwin")
SOLARIS = platform.startswith("sunos")
WINDOWS = platform.startswith("win32")
WINDOWS_XP = platform_lib.release() == "XP"
BSD = DARWIN or FREEBSD or OPENBSD or NETBSD
# See https://docs.python.org/3/library/platform.html#cross-platform
IS_64BITS = maxsize > 2**32
# LOOPBACK_NAME moved to conf.loopback_name

from ctypes import sizeof

from pocsuite3.thirdparty.scapy.structures import bpf_program
from pocsuite3.thirdparty.scapy.data import MTU
SIOCGIFFLAGS = 0xc0206911
BPF_BUFFER_LENGTH = MTU

# From net/bpf.h
BIOCIMMEDIATE = 0x80044270
BIOCGSTATS = 0x4008426f
BIOCPROMISC = 0x20004269
BIOCSETIF = 0x8020426c
BIOCSBLEN = 0xc0044266
BIOCGBLEN = 0x40044266
BIOCSETF = 0x80004267 | ((sizeof(bpf_program) & 0x1fff) << 16)
BIOCSDLT = 0x80044278
BIOCSHDRCMPLT = 0x80044275
BIOCGDLT = 0x4004426a
DLT_IEEE802_11_RADIO = 127



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

"""
Linux specific functions.
"""

# from __future__ import absolute_import


import array
from fcntl import ioctl
import os
from select import select
import socket
import struct
import time

import subprocess

from pocsuite3.thirdparty.scapy import consts
from pocsuite3.thirdparty.scapy.compat import raw, plain_str
from pocsuite3.thirdparty.scapy.consts import LINUX
from pocsuite3.thirdparty.scapy import utils
from pocsuite3.thirdparty.scapy import utils6
from pocsuite3.thirdparty.scapy.packet import Packet, Padding
from pocsuite3.thirdparty.scapy.config import conf
from pocsuite3.thirdparty.scapy.data import MTU, ETH_P_ALL, SOL_PACKET, SO_ATTACH_FILTER, \
    SO_TIMESTAMPNS
from pocsuite3.thirdparty.scapy.supersocket import SuperSocket
from pocsuite3.thirdparty.scapy.error import warning, Scapy_Exception, \
    ScapyInvalidPlatformException, log_runtime
from pocsuite3.thirdparty.scapy.common import get_if, compile_filter
import pocsuite3.thirdparty.scapy.six as six
from pocsuite3.thirdparty.scapy.six.moves import range

from pocsuite3.thirdparty.scapy.common import get_if_raw_hwaddr  # noqa: F401

# From bits/ioctls.h
SIOCGIFHWADDR = 0x8927          # Get hardware address
SIOCGIFADDR = 0x8915          # get PA address
SIOCGIFNETMASK = 0x891b          # get network PA mask
SIOCGIFNAME = 0x8910          # get iface name
SIOCSIFLINK = 0x8911          # set iface channel
SIOCGIFCONF = 0x8912          # get iface list
SIOCGIFFLAGS = 0x8913          # get flags
SIOCSIFFLAGS = 0x8914          # set flags
SIOCGIFINDEX = 0x8933          # name -> if_index mapping
SIOCGIFCOUNT = 0x8938          # get number of devices
SIOCGSTAMP = 0x8906          # get packet timestamp (as a timeval)

# From if.h
IFF_UP = 0x1               # Interface is up.
IFF_BROADCAST = 0x2        # Broadcast address valid.
IFF_DEBUG = 0x4            # Turn on debugging.
IFF_LOOPBACK = 0x8         # Is a loopback net.
IFF_POINTOPOINT = 0x10     # Interface is point-to-point link.
IFF_NOTRAILERS = 0x20      # Avoid use of trailers.
IFF_RUNNING = 0x40         # Resources allocated.
IFF_NOARP = 0x80           # No address resolution protocol.
IFF_PROMISC = 0x100        # Receive all packets.

# From netpacket/packet.h
PACKET_ADD_MEMBERSHIP = 1
PACKET_DROP_MEMBERSHIP = 2
PACKET_RECV_OUTPUT = 3
PACKET_RX_RING = 5
PACKET_STATISTICS = 6
PACKET_MR_MULTICAST = 0
PACKET_MR_PROMISC = 1
PACKET_MR_ALLMULTI = 2

# From net/route.h
RTF_UP = 0x0001  # Route usable
RTF_REJECT = 0x0200

# From if_packet.h
PACKET_HOST = 0  # To us
PACKET_BROADCAST = 1  # To all
PACKET_MULTICAST = 2  # To group
PACKET_OTHERHOST = 3  # To someone else
PACKET_OUTGOING = 4  # Outgoing of any type
PACKET_LOOPBACK = 5  # MC/BRD frame looped back
PACKET_USER = 6  # To user space
PACKET_KERNEL = 7  # To kernel space
PACKET_AUXDATA = 8
PACKET_FASTROUTE = 6  # Fastrouted frame
# Unused, PACKET_FASTROUTE and PACKET_LOOPBACK are invisible to user space

# Utils


def get_if_raw_addr(iff):
    try:
        return get_if(iff, SIOCGIFADDR)[20:24]
    except IOError:
        return b"\0\0\0\0"


def get_if_list():
    try:
        f = open("/proc/net/dev", "rb")
    except IOError:
        try:
            f.close()
        except Exception:
            pass
        warning("Can't open /proc/net/dev !")
        return []
    lst = []
    f.readline()
    f.readline()
    for line in f:
        line = plain_str(line)
        lst.append(line.split(":")[0].strip())
    f.close()
    return lst


def get_working_if():
    """
    Return the name of the first network interfcace that is up.
    """
    for i in get_if_list():
        if i == conf.loopback_name:
            continue
        ifflags = struct.unpack("16xH14x", get_if(i, SIOCGIFFLAGS))[0]
        if ifflags & IFF_UP:
            return i
    return conf.loopback_name


def attach_filter(sock, bpf_filter, iface):
    """
    Compile bpf filter and attach it to a socket

    :param sock: the python socket
    :param bpf_filter: the bpf string filter to compile
    :param iface: the interface used to compile
    """
    bp = compile_filter(bpf_filter, iface)
    sock.setsockopt(socket.SOL_SOCKET, SO_ATTACH_FILTER, bp)


def set_promisc(s, iff, val=1):
    mreq = struct.pack("IHH8s", get_if_index(iff), PACKET_MR_PROMISC, 0, b"")
    if val:
        cmd = PACKET_ADD_MEMBERSHIP
    else:
        cmd = PACKET_DROP_MEMBERSHIP
    s.setsockopt(SOL_PACKET, cmd, mreq)


def get_alias_address(iface_name, ip_mask, gw_str, metric):
    """
    Get the correct source IP address of an interface alias
    """

    # Detect the architecture
    if consts.IS_64BITS:
        offset, name_len = 16, 40
    else:
        offset, name_len = 32, 32

    # Retrieve interfaces structures
    sck = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    names = array.array('B', b'\0' * 4096)
    ifreq = ioctl(sck.fileno(), SIOCGIFCONF,
                  struct.pack("iL", len(names), names.buffer_info()[0]))

    # Extract interfaces names
    out = struct.unpack("iL", ifreq)[0]
    names = names.tobytes() if six.PY3 else names.tostring()
    names = [names[i:i + offset].split(b'\0', 1)[0] for i in range(0, out, name_len)]  # noqa: E501

    # Look for the IP address
    for ifname in names:
        # Only look for a matching interface name
        if not ifname.decode("utf8").startswith(iface_name):
            continue

        # Retrieve and convert addresses
        ifreq = ioctl(sck, SIOCGIFADDR, struct.pack("16s16x", ifname))
        ifaddr = struct.unpack(">I", ifreq[20:24])[0]
        ifreq = ioctl(sck, SIOCGIFNETMASK, struct.pack("16s16x", ifname))
        msk = struct.unpack(">I", ifreq[20:24])[0]

        # Get the full interface name
        ifname = plain_str(ifname)
        if ':' in ifname:
            ifname = ifname[:ifname.index(':')]
        else:
            continue

        # Check if the source address is included in the network
        if (ifaddr & msk) == ip_mask:
            sck.close()
            return (ifaddr & msk, msk, gw_str, ifname,
                    utils.ltoa(ifaddr), metric)

    sck.close()
    return


def read_routes():
    try:
        f = open("/proc/net/route", "rb")
    except IOError:
        warning("Can't open /proc/net/route !")
        return []
    routes = []
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        ifreq = ioctl(s, SIOCGIFADDR, struct.pack("16s16x", conf.loopback_name.encode("utf8")))  # noqa: E501
        addrfamily = struct.unpack("h", ifreq[16:18])[0]
        if addrfamily == socket.AF_INET:
            ifreq2 = ioctl(s, SIOCGIFNETMASK, struct.pack("16s16x", conf.loopback_name.encode("utf8")))  # noqa: E501
            msk = socket.ntohl(struct.unpack("I", ifreq2[20:24])[0])
            dst = socket.ntohl(struct.unpack("I", ifreq[20:24])[0]) & msk
            ifaddr = utils.inet_ntoa(ifreq[20:24])
            routes.append((dst, msk, "0.0.0.0", conf.loopback_name, ifaddr, 1))  # noqa: E501
        else:
            warning("Interface %s: unknown address family (%i)" % (conf.loopback_name, addrfamily))  # noqa: E501
    except IOError as err:
        if err.errno == 99:
            warning("Interface %s: no address assigned" % conf.loopback_name)  # noqa: E501
        else:
            warning("Interface %s: failed to get address config (%s)" % (conf.loopback_name, str(err)))  # noqa: E501

    for line in f.readlines()[1:]:
        line = plain_str(line)
        iff, dst, gw, flags, _, _, metric, msk, _, _, _ = line.split()
        flags = int(flags, 16)
        if flags & RTF_UP == 0:
            continue
        if flags & RTF_REJECT:
            continue
        try:
            ifreq = ioctl(s, SIOCGIFADDR, struct.pack("16s16x", iff.encode("utf8")))  # noqa: E501
        except IOError:  # interface is present in routing tables but does not have any assigned IP  # noqa: E501
            ifaddr = "0.0.0.0"
            ifaddr_int = 0
        else:
            addrfamily = struct.unpack("h", ifreq[16:18])[0]
            if addrfamily == socket.AF_INET:
                ifaddr = utils.inet_ntoa(ifreq[20:24])
                ifaddr_int = struct.unpack("!I", ifreq[20:24])[0]
            else:
                warning("Interface %s: unknown address family (%i)", iff, addrfamily)  # noqa: E501
                continue

        # Attempt to detect an interface alias based on addresses inconsistencies  # noqa: E501
        dst_int = socket.htonl(int(dst, 16)) & 0xffffffff
        msk_int = socket.htonl(int(msk, 16)) & 0xffffffff
        gw_str = utils.inet_ntoa(struct.pack("I", int(gw, 16)))
        metric = int(metric)

        if ifaddr_int & msk_int != dst_int:
            tmp_route = get_alias_address(iff, dst_int, gw_str, metric)
            if tmp_route:
                routes.append(tmp_route)
            else:
                routes.append((dst_int, msk_int, gw_str, iff, ifaddr, metric))

        else:
            routes.append((dst_int, msk_int, gw_str, iff, ifaddr, metric))

    f.close()
    s.close()
    return routes

############
#   IPv6   #
############


def in6_getifaddr():
    """
    Returns a list of 3-tuples of the form (addr, scope, iface) where
    'addr' is the address of scope 'scope' associated to the interface
    'iface'.

    This is the list of all addresses of all interfaces available on
    the system.
    """
    ret = []
    try:
        fdesc = open("/proc/net/if_inet6", "rb")
    except IOError:
        return ret
    for line in fdesc:
        # addr, index, plen, scope, flags, ifname
        tmp = plain_str(line).split()
        addr = utils6.in6_ptop(
            b':'.join(
                struct.unpack('4s4s4s4s4s4s4s4s', tmp[0].encode())
            ).decode()
        )
        # (addr, scope, iface)
        ret.append((addr, int(tmp[3], 16), tmp[5]))
    fdesc.close()
    return ret


def read_routes6():
    try:
        f = open("/proc/net/ipv6_route", "rb")
    except IOError:
        return []
    # 1. destination network
    # 2. destination prefix length
    # 3. source network displayed
    # 4. source prefix length
    # 5. next hop
    # 6. metric
    # 7. reference counter (?!?)
    # 8. use counter (?!?)
    # 9. flags
    # 10. device name
    routes = []

    def proc2r(p):
        ret = struct.unpack('4s4s4s4s4s4s4s4s', p)
        ret = b':'.join(ret).decode()
        return utils6.in6_ptop(ret)

    lifaddr = in6_getifaddr()
    for line in f.readlines():
        d, dp, _, _, nh, metric, rc, us, fl, dev = line.split()
        metric = int(metric, 16)
        fl = int(fl, 16)
        dev = plain_str(dev)

        if fl & RTF_UP == 0:
            continue
        if fl & RTF_REJECT:
            continue

        d = proc2r(d)
        dp = int(dp, 16)
        nh = proc2r(nh)

        cset = []  # candidate set (possible source addresses)
        if dev == conf.loopback_name:
            if d == '::':
                continue
            cset = ['::1']
        else:
            devaddrs = (x for x in lifaddr if x[2] == dev)
            cset = utils6.construct_source_candidate_set(d, dp, devaddrs)

        if len(cset) != 0:
            routes.append((d, dp, nh, dev, cset, metric))
    f.close()
    return routes


def get_if_index(iff):
    return int(struct.unpack("I", get_if(iff, SIOCGIFINDEX)[16:20])[0])


if os.uname()[4] in ['x86_64', 'aarch64']:
    def get_last_packet_timestamp(sock):
        ts = ioctl(sock, SIOCGSTAMP, "1234567890123456")
        s, us = struct.unpack("QQ", ts)
        return s + us / 1000000.0
else:
    def get_last_packet_timestamp(sock):
        ts = ioctl(sock, SIOCGSTAMP, "12345678")
        s, us = struct.unpack("II", ts)
        return s + us / 1000000.0


def _flush_fd(fd):
    if hasattr(fd, 'fileno'):
        fd = fd.fileno()
    while True:
        r, w, e = select([fd], [], [], 0)
        if r:
            os.read(fd, MTU)
        else:
            break


class L2Socket(SuperSocket):
    desc = "read/write packets at layer 2 using Linux PF_PACKET sockets"

    def __init__(self, iface=None, type=ETH_P_ALL, promisc=None, filter=None,
                 nofilter=0, monitor=None):
        self.iface = conf.iface if iface is None else iface
        self.type = type
        self.promisc = conf.sniff_promisc if promisc is None else promisc
        if monitor is not None:
            log_runtime.info(
                "The 'monitor' argument has no effect on native linux sockets."
            )
        self.ins = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.htons(type))  # noqa: E501
        if not nofilter:
            if conf.except_filter:
                if filter:
                    filter = "(%s) and not (%s)" % (filter, conf.except_filter)
                else:
                    filter = "not (%s)" % conf.except_filter
            if filter is not None:
                try:
                    attach_filter(self.ins, filter, iface)
                except ImportError as ex:
                    warning("Cannot set filter: %s" % ex)
        if self.promisc:
            set_promisc(self.ins, self.iface)
        self.ins.bind((self.iface, type))
        _flush_fd(self.ins)
        self.ins.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_RCVBUF,
            conf.bufsize
        )

        # Receive Auxiliary Data (VLAN tags)
        try:
            self.ins.setsockopt(SOL_PACKET, PACKET_AUXDATA, 1)
            self.ins.setsockopt(
                socket.SOL_SOCKET,
                SO_TIMESTAMPNS,
                1
            )
            self.auxdata_available = True
        except OSError:
            # Note: Auxiliary Data is only supported since
            #       Linux 2.6.21
            msg = "Your Linux Kernel does not support Auxiliary Data!"
            log_runtime.info(msg)
        if isinstance(self, L2ListenSocket):
            self.outs = None
        else:
            self.outs = self.ins
            self.outs.setsockopt(
                socket.SOL_SOCKET,
                socket.SO_SNDBUF,
                conf.bufsize
            )
        sa_ll = self.ins.getsockname()
        if sa_ll[3] in conf.l2types:
            self.LL = conf.l2types[sa_ll[3]]
            self.lvl = 2
        elif sa_ll[1] in conf.l3types:
            self.LL = conf.l3types[sa_ll[1]]
            self.lvl = 3
        else:
            self.LL = conf.default_l2
            self.lvl = 2
            warning("Unable to guess type (interface=%s protocol=%#x family=%i). Using %s", sa_ll[0], sa_ll[1], sa_ll[3], self.LL.name)  # noqa: E501

    def close(self):
        if self.closed:
            return
        try:
            if self.promisc and self.ins:
                set_promisc(self.ins, self.iface, 0)
        except (AttributeError, OSError):
            pass
        SuperSocket.close(self)

    def recv_raw(self, x=MTU):
        """Receives a packet, then returns a tuple containing (cls, pkt_data, time)"""  # noqa: E501
        pkt, sa_ll, ts = self._recv_raw(self.ins, x)
        if self.outs and sa_ll[2] == socket.PACKET_OUTGOING:
            return None, None, None
        if ts is None:
            ts = get_last_packet_timestamp(self.ins)
        return self.LL, pkt, ts

    def send(self, x):
        try:
            return SuperSocket.send(self, x)
        except socket.error as msg:
            if msg.errno == 22 and len(x) < conf.min_pkt_size:
                padding = b"\x00" * (conf.min_pkt_size - len(x))
                if isinstance(x, Packet):
                    return SuperSocket.send(self, x / Padding(load=padding))
                else:
                    return SuperSocket.send(self, raw(x) + padding)
            raise


class L2ListenSocket(L2Socket):
    desc = "read packets at layer 2 using Linux PF_PACKET sockets. Also receives the packets going OUT"  # noqa: E501

    def send(self, x):
        raise Scapy_Exception("Can't send anything with L2ListenSocket")


class L3PacketSocket(L2Socket):
    desc = "read/write packets at layer 3 using Linux PF_PACKET sockets"

    def recv(self, x=MTU):
        pkt = SuperSocket.recv(self, x)
        if pkt and self.lvl == 2:
            pkt.payload.time = pkt.time
            return pkt.payload
        return pkt

    def send(self, x):
        iff = x.route()[0]
        if iff is None:
            iff = conf.iface
        sdto = (iff, self.type)
        self.outs.bind(sdto)
        sn = self.outs.getsockname()
        ll = lambda x: x
        type_x = type(x)
        if type_x in conf.l3types:
            sdto = (iff, conf.l3types[type_x])
        if sn[3] in conf.l2types:
            ll = lambda x: conf.l2types[sn[3]]() / x
        if self.lvl == 3 and type_x != self.LL:
            warning("Incompatible L3 types detected using %s instead of %s !",
                    type_x, self.LL)
            self.LL = type_x
        sx = raw(ll(x))
        x.sent_time = time.time()
        try:
            self.outs.sendto(sx, sdto)
        except socket.error as msg:
            if msg.errno == 22 and len(sx) < conf.min_pkt_size:
                self.outs.send(sx + b"\x00" * (conf.min_pkt_size - len(sx)))
            elif conf.auto_fragment and msg.errno == 90:
                for p in x.fragment():
                    self.outs.sendto(raw(ll(p)), sdto)
            else:
                raise


class VEthPair(object):
    """
    encapsulates a virtual Ethernet interface pair
    """

    def __init__(self, iface_name, peer_name):

        if not LINUX:
            # ToDo: do we need a kernel version check here?
            raise ScapyInvalidPlatformException(
                'Virtual Ethernet interface pair only available on Linux'
            )

        self.ifaces = [iface_name, peer_name]

    def iface(self):
        return self.ifaces[0]

    def peer(self):
        return self.ifaces[1]

    def setup(self):
        """
        create veth pair links
        :raises subprocess.CalledProcessError if operation fails
        """
        subprocess.check_call(['ip', 'link', 'add', self.ifaces[0], 'type', 'veth', 'peer', 'name', self.ifaces[1]])  # noqa: E501

    def destroy(self):
        """
        remove veth pair links
        :raises subprocess.CalledProcessError if operation fails
        """
        subprocess.check_call(['ip', 'link', 'del', self.ifaces[0]])

    def up(self):
        """
        set veth pair links up
        :raises subprocess.CalledProcessError if operation fails
        """
        for idx in [0, 1]:
            subprocess.check_call(["ip", "link", "set", self.ifaces[idx], "up"])  # noqa: E501

    def down(self):
        """
        set veth pair links down
        :raises subprocess.CalledProcessError if operation fails
        """
        for idx in [0, 1]:
            subprocess.check_call(["ip", "link", "set", self.ifaces[idx], "down"])  # noqa: E501

    def __enter__(self):
        self.setup()
        self.up()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.destroy()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2018/12/25 上午10:49
# @Author  : chenghs
# @File    : console.py
import os
import sys

try:
    import pocsuite3
except ImportError:
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir)))
from pocsuite3.cli import check_environment, module_path
from pocsuite3 import set_paths
from pocsuite3.lib.core.interpreter import PocsuiteInterpreter
from pocsuite3.lib.core.option import init_options


def main():
    check_environment()
    set_paths(module_path())
    init_options()
    poc = PocsuiteInterpreter()
    poc.start()


if __name__ == '__main__':
    main()



import re
from pocsuite3.lib.core.data import conf
from pocsuite3.lib.core.data import logger

def regex_rule(files):
    if not conf.rule_filename:
        conf.rule_filename = "rule.rule"
    for file_name in files:
        regx_rules = ["name = '(.*)'",
                      "suricata_request = '''([\s\S]*?)'''",
                      "references = \['(.*)'\]", "createDate = '(.*)'", "updateDate = '(.*)'",
                      "vulID = '(.*)'",
                      "version = '(.*)'",
                      "suricata_response = '''([\s\S]*?)'''",
                      ]

        information_list = {"name": "0",
                            "suricata_request": "1",
                            "references": "2",
                            "createDate": "3",
                            "updateDate": "4",
                            "vulID": "5",
                            "version": "6",
                            "suricata_response": "7",
                            "flowbits": ""}

        f = open(file_name, "r", encoding="utf-8")
        st = f.read()
        for key, value in information_list.items():
            if value:
                pattern = re.compile(regx_rules[int(value)])
                cve_list = pattern.findall(st)
                if cve_list:
                    if "name" in regx_rules[int(value)]:
                        information_list[key] = cve_list[0].replace("\n", "")
                    else:
                        if "suricata_request" not in regx_rules[int(value)] and "suricata_response" not in regx_rules[int(value)]:
                            information_list[key] = cve_list[0].replace("\n", "").replace(" ", "")
                        else:
                            information_list[key] = cve_list[0].replace("\n", "")
                else:
                    information_list[key] = ""
        if not information_list["suricata_request"]:
            continue
        if "、" in information_list["vulID"]:
            information_list["vulID"] = information_list["vulID"].split("、")[0]
        elif not information_list["vulID"]:
            information_list["vulID"] = 0
        if information_list["suricata_response"] and not conf.rule_req:
            # 6220553==seebug.(　ˇωˇ)
            rule_to_server = '''alert http any any -> any any (msg:"{}";flow:established,to_server;{}classtype:web-application-attack;reference:url,{}; metadata:created_at {}, updated_at {};flowbits:set,{};flowbits:noalert;sid:{};rev:{};)'''.format(
                information_list["name"], information_list["suricata_request"], information_list["references"],
                information_list["createDate"], information_list["updateDate"], information_list["name"].replace(" ", "_"),
                6220553 + int(float(information_list["vulID"])) * 2, int(float(information_list["version"])))

            rule_to_client = '''alert http any any -> any any (msg:"{}";flow:established,to_client;{}classtype:web-application-attack;reference:url,{}; metadata:created_at {}, updated_at {};flowbits:isset,{};sid:{};rev:{};)'''.format(
                information_list["name"], information_list["suricata_response"], information_list["references"],
                information_list["createDate"], information_list["updateDate"], information_list["name"].replace(" ", "_"),
                6220553 + int(float(information_list["vulID"])) * 2 + 1, int(float(information_list["version"])))
        else:
            rule_to_server = '''alert http any any -> any any (msg:"{}";flow:established,to_server;{}classtype:web-application-attack;reference:url,{}; metadata:created_at {}, updated_at {};sid:{};rev:{};)'''.format(
                information_list["name"], information_list["suricata_request"], information_list["references"],
                information_list["createDate"], information_list["updateDate"],
                6220553 + int(float(information_list["vulID"])) * 2,
                int(float(information_list["version"])))
            rule_to_client = ""
        with open(conf.rule_filename, "a", encoding="utf-8") as f:
            f.write(rule_to_server+"\n")
            f.write(rule_to_client+"\n")
        f.close()
        logger.info("{} rule is:".format(file_name[file_name.rfind("\\")+1:]))
        print(rule_to_server)
        print(rule_to_client)





class ShellCode:
    def __init__(self, os_target='', os_target_arch='', connect_back_ip='localhost', connect_back_port=5555,
                 bad_chars=[], prefix='', suffix=''):
        self.os_target = os_target
        self.os_target_arch = os_target_arch
        self.connect_back_ip = connect_back_ip
        self.connect_back_port = connect_back_port
        self.bad_chars = bad_chars
        self.prefix = prefix
        self.suffix = suffix
        self.name = ''

    def format_shellcode(self, code):
        if isinstance(code, str):
            code = code.replace('{{LOCALHOST}}', self.connect_back_ip)
            code = code.replace('{{LOCALPORT}}', str(self.connect_back_port))

        return code

    def get_shellcode(self, inline=False):
        return ''

    def make_inline(self, payload):
        payload = payload.replace('\t', ' ')
        payload = payload.replace('\r', ' ')
        payload = payload.replace('\n', ' ')
        return payload



import urllib

from configparser import ConfigParser
from pocsuite3.lib.core.data import logger
from pocsuite3.lib.core.data import paths
from pocsuite3.lib.request import requests


class Shodan():
    def __init__(self, conf_path=paths.POCSUITE_RC_PATH, token=None):
        self.headers = None
        self.credits = 0
        self.conf_path = conf_path

        if self.conf_path:
            self.parser = ConfigParser()
            self.parser.read(self.conf_path)
            try:
                self.token = self.parser.get("Shodan", 'Token')
            except Exception:
                pass

        if token:
            self.token = token
        self.check_token()

    def token_is_available(self):
        if self.token:
            try:
                resp = requests.get('https://api.shodan.io/account/profile?key={0}'.format(self.token))
                if resp and resp.status_code == 200 and "member" in resp.json():
                    return True
            except Exception as ex:
                logger.error(str(ex))
        return False

    def check_token(self):
        if self.token_is_available():
            return True
        else:

            new_token = input("Shodan API Token:")
            self.token = new_token
            if self.token_is_available():
                self.write_conf()
                return True
            else:
                logger.error("The shodan api token is incorrect. "
                             "Please enter the correct api token.")
                self.check_token()

    def write_conf(self):
        if not self.parser.has_section("Shodan"):
            self.parser.add_section("Shodan")
        try:
            self.parser.set("Shodan", "Token", self.token)
            self.parser.write(open(self.conf_path, "w"))
        except Exception as ex:
            logger.error(str(ex))

    def get_resource_info(self):
        try:
            resp = requests.get('https://api.shodan.io/account/profile?key={0}'.format(self.token))
            if resp and resp.status_code == 200 and 'credits' in resp.json():
                content = resp.json()
                self.credits = content['credits']
                return True
        except Exception as ex:
            logger.error(str(ex))
        return False

    def search(self, dork, pages=1, resource='host'):
        search_result = set()
        try:
            for page in range(1, pages + 1):
                url = "https://api.shodan.io/shodan/{0}/search?key={1}&query={2}&page={3}".format(resource,
                                                                                                  self.token,
                                                                                                  urllib.parse.quote(
                                                                                                      dork),
                                                                                                  page)
                resp = requests.get(url)
                if resp and resp.status_code == 200 and "total" in resp.json():
                    content = resp.json()
                    for match in content['matches']:
                        ans = match['ip_str']
                        if 'port' in match:
                            ans += ':' + str(match['port'])
                        search_result.add(ans)
                else:
                    logger.error("[PLUGIN] Shodan:{}".format(resp.text))
        except Exception as ex:
            logger.error(str(ex))
        return search_result


if __name__ == "__main__":
    sd = Shodan()
    sd.search('dedecms')



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

"""
Generators and packet meta classes.
"""

################
#  Generators  #
################

# from __future__ import absolute_import

from functools import reduce
import operator
import os
import re
import random
import socket
import subprocess
import types

from pocsuite3.thirdparty.scapy.consts import WINDOWS
from pocsuite3.thirdparty.scapy.six.moves import range


class Gen(object):
    __slots__ = []

    def __iter__(self):
        return iter([])

    def __iterlen__(self):
        return sum(1 for _ in iter(self))


def _get_values(value):
    """Generate a range object from (start, stop[, step]) tuples, or
    return value.

    """
    if (isinstance(value, tuple) and (2 <= len(value) <= 3) and
            all(hasattr(i, "__int__") for i in value)):
        # We use values[1] + 1 as stop value for (x)range to maintain
        # the behavior of using tuples as field `values`
        return range(*((int(value[0]), int(value[1]) + 1) +
                       tuple(int(v) for v in value[2:])))
    return value


class SetGen(Gen):
    def __init__(self, values, _iterpacket=1):
        self._iterpacket = _iterpacket
        if isinstance(values, (list, BasePacketList)):
            self.values = [_get_values(val) for val in values]
        else:
            self.values = [_get_values(values)]

    def transf(self, element):
        return element

    def __iter__(self):
        for i in self.values:
            if (isinstance(i, Gen) and
                (self._iterpacket or not isinstance(i, BasePacket))) or (
                    isinstance(i, (range, types.GeneratorType))):
                for j in i:
                    yield j
            else:
                yield i

    def __repr__(self):
        return "<SetGen %r>" % self.values


class Net(Gen):
    """Generate a list of IPs from a network address or a name"""
    name = "ip"
    ip_regex = re.compile(r"^(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)\.(\*|[0-2]?[0-9]?[0-9](-[0-2]?[0-9]?[0-9])?)(/[0-3]?[0-9])?$")  # noqa: E501

    @staticmethod
    def _parse_digit(a, netmask):
        netmask = min(8, max(netmask, 0))
        if a == "*":
            a = (0, 256)
        elif a.find("-") >= 0:
            x, y = [int(d) for d in a.split('-')]
            if x > y:
                y = x
            a = (x & (0xff << netmask), max(y, (x | (0xff >> (8 - netmask)))) + 1)  # noqa: E501
        else:
            a = (int(a) & (0xff << netmask), (int(a) | (0xff >> (8 - netmask))) + 1)  # noqa: E501
        return a

    @classmethod
    def _parse_net(cls, net):
        tmp = net.split('/') + ["32"]
        if not cls.ip_regex.match(net):
            tmp[0] = socket.gethostbyname(tmp[0])
        netmask = int(tmp[1])
        ret_list = [cls._parse_digit(x, y - netmask) for (x, y) in zip(tmp[0].split('.'), [8, 16, 24, 32])]  # noqa: E501
        return ret_list, netmask

    def __init__(self, net):
        self.repr = net
        self.parsed, self.netmask = self._parse_net(net)

    def __str__(self):
        return next(self.__iter__(), None)

    def __iter__(self):
        for d in range(*self.parsed[3]):
            for c in range(*self.parsed[2]):
                for b in range(*self.parsed[1]):
                    for a in range(*self.parsed[0]):
                        yield "%i.%i.%i.%i" % (a, b, c, d)

    def __iterlen__(self):
        return reduce(operator.mul, ((y - x) for (x, y) in self.parsed), 1)

    def choice(self):
        return ".".join(str(random.randint(v[0], v[1] - 1)) for v in self.parsed)  # noqa: E501

    def __repr__(self):
        return "Net(%r)" % self.repr

    def __eq__(self, other):
        if not other:
            return False
        if hasattr(other, "parsed"):
            p2 = other.parsed
        else:
            p2, nm2 = self._parse_net(other)
        return self.parsed == p2

    def __ne__(self, other):
        # Python 2.7 compat
        return not self == other

    __hash__ = None

    def __contains__(self, other):
        if hasattr(other, "parsed"):
            p2 = other.parsed
        else:
            p2, nm2 = self._parse_net(other)
        return all(a1 <= a2 and b1 >= b2 for (a1, b1), (a2, b2) in zip(self.parsed, p2))  # noqa: E501

    def __rcontains__(self, other):
        return self in self.__class__(other)


class OID(Gen):
    name = "OID"

    def __init__(self, oid):
        self.oid = oid
        self.cmpt = []
        fmt = []
        for i in oid.split("."):
            if "-" in i:
                fmt.append("%i")
                self.cmpt.append(tuple(map(int, i.split("-"))))
            else:
                fmt.append(i)
        self.fmt = ".".join(fmt)

    def __repr__(self):
        return "OID(%r)" % self.oid

    def __iter__(self):
        ii = [k[0] for k in self.cmpt]
        while True:
            yield self.fmt % tuple(ii)
            i = 0
            while True:
                if i >= len(ii):
                    return
                if ii[i] < self.cmpt[i][1]:
                    ii[i] += 1
                    break
                else:
                    ii[i] = self.cmpt[i][0]
                i += 1

    def __iterlen__(self):
        return reduce(operator.mul, (max(y - x, 0) + 1 for (x, y) in self.cmpt), 1)  # noqa: E501


######################################
#  Packet abstract and base classes  #
######################################

class Packet_metaclass(type):
    def __new__(cls, name, bases, dct):
        if "fields_desc" in dct:  # perform resolution of references to other packets  # noqa: E501
            current_fld = dct["fields_desc"]
            resolved_fld = []
            for f in current_fld:
                if isinstance(f, Packet_metaclass):  # reference to another fields_desc  # noqa: E501
                    for f2 in f.fields_desc:
                        resolved_fld.append(f2)
                else:
                    resolved_fld.append(f)
        else:  # look for a fields_desc in parent classes
            resolved_fld = None
            for b in bases:
                if hasattr(b, "fields_desc"):
                    resolved_fld = b.fields_desc
                    break

        if resolved_fld:  # perform default value replacements
            final_fld = []
            for f in resolved_fld:
                if f.name in dct:
                    f = f.copy()
                    f.default = dct[f.name]
                    del(dct[f.name])
                final_fld.append(f)

            dct["fields_desc"] = final_fld

        dct.setdefault("__slots__", [])
        for attr in ["name", "overload_fields"]:
            try:
                dct["_%s" % attr] = dct.pop(attr)
            except KeyError:
                pass
        newcls = super(Packet_metaclass, cls).__new__(cls, name, bases, dct)
        newcls.__all_slots__ = set(
            attr
            for cls in newcls.__mro__ if hasattr(cls, "__slots__")
            for attr in cls.__slots__
        )

        newcls.aliastypes = [newcls] + getattr(newcls, "aliastypes", [])

        if hasattr(newcls, "register_variant"):
            newcls.register_variant()
        for f in newcls.fields_desc:
            if hasattr(f, "register_owner"):
                f.register_owner(newcls)
        if newcls.__name__[0] != "_":
            from pocsuite3.thirdparty.scapy import config
            config.conf.layers.register(newcls)
        return newcls

    def __getattr__(self, attr):
        for k in self.fields_desc:
            if k.name == attr:
                return k
        raise AttributeError(attr)

    def __call__(cls, *args, **kargs):
        if "dispatch_hook" in cls.__dict__:
            try:
                cls = cls.dispatch_hook(*args, **kargs)
            except Exception:
                from pocsuite3.thirdparty.scapy import config
                if config.conf.debug_dissector:
                    raise
                cls = config.conf.raw_layer
        i = cls.__new__(cls, cls.__name__, cls.__bases__, cls.__dict__)
        i.__init__(*args, **kargs)
        return i


class Field_metaclass(type):
    def __new__(cls, name, bases, dct):
        dct.setdefault("__slots__", [])
        newcls = super(Field_metaclass, cls).__new__(cls, name, bases, dct)
        return newcls


class BasePacket(Gen):
    __slots__ = []


#############################
#  Packet list base class   #
#############################

class BasePacketList(object):
    __slots__ = []


class _CanvasDumpExtended(object):
    def psdump(self, filename=None, **kargs):
        """
        psdump(filename=None, layer_shift=0, rebuild=1)

        Creates an EPS file describing a packet. If filename is not provided a
        temporary file is created and gs is called.

        :param filename: the file's filename
        """
        from pocsuite3.thirdparty.scapy.config import conf
        from pocsuite3.thirdparty.scapy.utils import get_temp_file, ContextManagerSubprocess
        canvas = self.canvas_dump(**kargs)
        if filename is None:
            fname = get_temp_file(autoext=kargs.get("suffix", ".eps"))
            canvas.writeEPSfile(fname)
            if WINDOWS and conf.prog.psreader is None:
                os.startfile(fname)
            else:
                with ContextManagerSubprocess(conf.prog.psreader):
                    subprocess.Popen([conf.prog.psreader, fname])
        else:
            canvas.writeEPSfile(filename)
        print()

    def pdfdump(self, filename=None, **kargs):
        """
        pdfdump(filename=None, layer_shift=0, rebuild=1)

        Creates a PDF file describing a packet. If filename is not provided a
        temporary file is created and xpdf is called.

        :param filename: the file's filename
        """
        from pocsuite3.thirdparty.scapy.config import conf
        from pocsuite3.thirdparty.scapy.utils import get_temp_file, ContextManagerSubprocess
        canvas = self.canvas_dump(**kargs)
        if filename is None:
            fname = get_temp_file(autoext=kargs.get("suffix", ".pdf"))
            canvas.writePDFfile(fname)
            if WINDOWS and conf.prog.pdfreader is None:
                os.startfile(fname)
            else:
                with ContextManagerSubprocess(conf.prog.pdfreader):
                    subprocess.Popen([conf.prog.pdfreader, fname])
        else:
            canvas.writePDFfile(filename)
        print()

    def svgdump(self, filename=None, **kargs):
        """
        svgdump(filename=None, layer_shift=0, rebuild=1)

        Creates an SVG file describing a packet. If filename is not provided a
        temporary file is created and gs is called.

        :param filename: the file's filename
        """
        from pocsuite3.thirdparty.scapy.config import conf
        from pocsuite3.thirdparty.scapy.utils import get_temp_file, ContextManagerSubprocess
        canvas = self.canvas_dump(**kargs)
        if filename is None:
            fname = get_temp_file(autoext=kargs.get("suffix", ".svg"))
            canvas.writeSVGfile(fname)
            if WINDOWS and conf.prog.svgreader is None:
                os.startfile(fname)
            else:
                with ContextManagerSubprocess(conf.prog.svgreader):
                    subprocess.Popen([conf.prog.svgreader, fname])
        else:
            canvas.writeSVGfile(filename)
        print()



# load poc from pocs directories
import os

from pocsuite3.api import PLUGIN_TYPE
from pocsuite3.api import PluginBase
from pocsuite3.api import logger
from pocsuite3.api import paths
from pocsuite3.api import register_plugin


class PocFromPocs(PluginBase):
    category = PLUGIN_TYPE.POCS

    def init(self):
        _pocs = []
        for root, dirs, files in os.walk(paths.POCSUITE_POCS_PATH):
            files = filter(lambda x: not x.startswith("__") and x.endswith(".py"), files)
            _pocs.extend(map(lambda x: os.path.join(root, x), files))
        for f in _pocs:

            if self.add_poc_from_file(f):
                info_msg = "[PLUGIN] load PoC script '{0}' from pocs directories success".format(f)
            else:
                info_msg = "[PLUGIN] load PoC script '{0}' from pocs directories failed".format(f)
            logger.info(info_msg)


register_plugin(PocFromPocs)



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

# Copyright (C) 2005  Guillaume Valadon <guedou@hongo.wide.ad.jp>
#                     Arnaud Ebalard <arnaud.ebalard@eads.net>

"""
Utility functions for IPv6.
"""
# from __future__ import absolute_import
import operator
import random
import socket
import struct
import time
import re

from pocsuite3.thirdparty.scapy.config import conf
from pocsuite3.thirdparty.scapy.base_classes import Gen
from pocsuite3.thirdparty.scapy.data import IPV6_ADDR_GLOBAL, IPV6_ADDR_LINKLOCAL, \
    IPV6_ADDR_SITELOCAL, IPV6_ADDR_LOOPBACK, IPV6_ADDR_UNICAST,\
    IPV6_ADDR_MULTICAST, IPV6_ADDR_6TO4, IPV6_ADDR_UNSPECIFIED
from pocsuite3.thirdparty.scapy.utils import *
from pocsuite3.thirdparty.scapy.compat import orb, chb
from pocsuite3.thirdparty.scapy.pton_ntop import inet_pton, inet_ntop
from pocsuite3.thirdparty.scapy.volatile import RandMAC
from pocsuite3.thirdparty.scapy.error import warning, Scapy_Exception
from functools import reduce, cmp_to_key
from pocsuite3.thirdparty.scapy.six.moves import range, zip


def construct_source_candidate_set(addr, plen, laddr):
    """
    Given all addresses assigned to a specific interface ('laddr' parameter),
    this function returns the "candidate set" associated with 'addr/plen'.

    Basically, the function filters all interface addresses to keep only those
    that have the same scope as provided prefix.

    This is on this list of addresses that the source selection mechanism
    will then be performed to select the best source address associated
    with some specific destination that uses this prefix.
    """
    def cset_sort(x, y):
        x_global = 0
        if in6_isgladdr(x):
            x_global = 1
        y_global = 0
        if in6_isgladdr(y):
            y_global = 1
        res = y_global - x_global
        if res != 0 or y_global != 1:
            return res
        # two global addresses: if one is native, it wins.
        if not in6_isaddr6to4(x):
            return -1
        return -res

    cset = []
    if in6_isgladdr(addr) or in6_isuladdr(addr):
        cset = (x for x in laddr if x[1] == IPV6_ADDR_GLOBAL)
    elif in6_islladdr(addr):
        cset = (x for x in laddr if x[1] == IPV6_ADDR_LINKLOCAL)
    elif in6_issladdr(addr):
        cset = (x for x in laddr if x[1] == IPV6_ADDR_SITELOCAL)
    elif in6_ismaddr(addr):
        if in6_ismnladdr(addr):
            cset = [('::1', 16, conf.loopback_name)]
        elif in6_ismgladdr(addr):
            cset = (x for x in laddr if x[1] == IPV6_ADDR_GLOBAL)
        elif in6_ismlladdr(addr):
            cset = (x for x in laddr if x[1] == IPV6_ADDR_LINKLOCAL)
        elif in6_ismsladdr(addr):
            cset = (x for x in laddr if x[1] == IPV6_ADDR_SITELOCAL)
    elif addr == '::' and plen == 0:
        cset = (x for x in laddr if x[1] == IPV6_ADDR_GLOBAL)
    cset = [x[0] for x in cset]
    # TODO convert the cmd use into a key
    cset.sort(key=cmp_to_key(cset_sort))  # Sort with global addresses first
    return cset


def get_source_addr_from_candidate_set(dst, candidate_set):
    """
    This function implement a limited version of source address selection
    algorithm defined in section 5 of RFC 3484. The format is very different
    from that described in the document because it operates on a set
    of candidate source address for some specific route.
    """

    def scope_cmp(a, b):
        """
        Given two addresses, returns -1, 0 or 1 based on comparison of
        their scope
        """
        scope_mapper = {IPV6_ADDR_GLOBAL: 4,
                        IPV6_ADDR_SITELOCAL: 3,
                        IPV6_ADDR_LINKLOCAL: 2,
                        IPV6_ADDR_LOOPBACK: 1}
        sa = in6_getscope(a)
        if sa == -1:
            sa = IPV6_ADDR_LOOPBACK
        sb = in6_getscope(b)
        if sb == -1:
            sb = IPV6_ADDR_LOOPBACK

        sa = scope_mapper[sa]
        sb = scope_mapper[sb]

        if sa == sb:
            return 0
        if sa > sb:
            return 1
        return -1

    def rfc3484_cmp(source_a, source_b):
        """
        The function implements a limited version of the rules from Source
        Address selection algorithm defined section of RFC 3484.
        """

        # Rule 1: Prefer same address
        if source_a == dst:
            return 1
        if source_b == dst:
            return 1

        # Rule 2: Prefer appropriate scope
        tmp = scope_cmp(source_a, source_b)
        if tmp == -1:
            if scope_cmp(source_a, dst) == -1:
                return 1
            else:
                return -1
        elif tmp == 1:
            if scope_cmp(source_b, dst) == -1:
                return 1
            else:
                return -1

        # Rule 3: cannot be easily implemented
        # Rule 4: cannot be easily implemented
        # Rule 5: does not make sense here
        # Rule 6: cannot be implemented
        # Rule 7: cannot be implemented

        # Rule 8: Longest prefix match
        tmp1 = in6_get_common_plen(source_a, dst)
        tmp2 = in6_get_common_plen(source_b, dst)
        if tmp1 > tmp2:
            return 1
        elif tmp2 > tmp1:
            return -1
        return 0

    if not candidate_set:
        # Should not happen
        return None

    candidate_set.sort(key=cmp_to_key(rfc3484_cmp), reverse=True)

    return candidate_set[0]


# Think before modify it : for instance, FE::1 does exist and is unicast
# there are many others like that.
# TODO : integrate Unique Local Addresses
def in6_getAddrType(addr):
    naddr = inet_pton(socket.AF_INET6, addr)
    paddr = inet_ntop(socket.AF_INET6, naddr)  # normalize
    addrType = 0
    # _Assignable_ Global Unicast Address space
    # is defined in RFC 3513 as those in 2000::/3
    if ((orb(naddr[0]) & 0xE0) == 0x20):
        addrType = (IPV6_ADDR_UNICAST | IPV6_ADDR_GLOBAL)
        if naddr[:2] == b' \x02':  # Mark 6to4 @
            addrType |= IPV6_ADDR_6TO4
    elif orb(naddr[0]) == 0xff:  # multicast
        addrScope = paddr[3]
        if addrScope == '2':
            addrType = (IPV6_ADDR_LINKLOCAL | IPV6_ADDR_MULTICAST)
        elif addrScope == 'e':
            addrType = (IPV6_ADDR_GLOBAL | IPV6_ADDR_MULTICAST)
        else:
            addrType = (IPV6_ADDR_GLOBAL | IPV6_ADDR_MULTICAST)
    elif ((orb(naddr[0]) == 0xfe) and ((int(paddr[2], 16) & 0xC) == 0x8)):
        addrType = (IPV6_ADDR_UNICAST | IPV6_ADDR_LINKLOCAL)
    elif paddr == "::1":
        addrType = IPV6_ADDR_LOOPBACK
    elif paddr == "::":
        addrType = IPV6_ADDR_UNSPECIFIED
    else:
        # Everything else is global unicast (RFC 3513)
        # Even old deprecated (RFC3879) Site-Local addresses
        addrType = (IPV6_ADDR_GLOBAL | IPV6_ADDR_UNICAST)

    return addrType


def in6_mactoifaceid(mac, ulbit=None):
    """
    Compute the interface ID in modified EUI-64 format associated
    to the Ethernet address provided as input.
    value taken by U/L bit in the interface identifier is basically
    the reversed value of that in given MAC address it can be forced
    to a specific value by using optional 'ulbit' parameter.
    """
    if len(mac) != 17:
        return None
    m = "".join(mac.split(':'))
    if len(m) != 12:
        return None
    first = int(m[0:2], 16)
    if ulbit is None or not (ulbit == 0 or ulbit == 1):
        ulbit = [1, '-', 0][first & 0x02]
    ulbit *= 2
    first = "%.02x" % ((first & 0xFD) | ulbit)
    eui64 = first + m[2:4] + ":" + m[4:6] + "FF:FE" + m[6:8] + ":" + m[8:12]
    return eui64.upper()


def in6_ifaceidtomac(ifaceid):
    """
    Extract the mac address from provided iface ID. Iface ID is provided
    in printable format ("XXXX:XXFF:FEXX:XXXX", eventually compressed). None
    is returned on error.
    """
    try:
        # Set ifaceid to a binary form
        ifaceid = inet_pton(socket.AF_INET6, "::" + ifaceid)[8:16]
    except Exception:
        return None
    if ifaceid[3:5] != b'\xff\xfe':  # Check for burned-in MAC address
        return None

    # Unpacking and converting first byte of faceid to MAC address equivalent
    first = struct.unpack("B", ifaceid[:1])[0]
    ulbit = 2 * [1, '-', 0][first & 0x02]
    first = struct.pack("B", ((first & 0xFD) | ulbit))
    # Split into two vars to remove the \xff\xfe bytes
    oui = first + ifaceid[1:3]
    end = ifaceid[5:]
    # Convert and reconstruct into a MAC Address
    mac_bytes = ["%.02x" % orb(x) for x in list(oui + end)]
    return ":".join(mac_bytes)


def in6_addrtomac(addr):
    """
    Extract the mac address from provided address. None is returned
    on error.
    """
    mask = inet_pton(socket.AF_INET6, "::ffff:ffff:ffff:ffff")
    x = in6_and(mask, inet_pton(socket.AF_INET6, addr))
    ifaceid = inet_ntop(socket.AF_INET6, x)[2:]
    return in6_ifaceidtomac(ifaceid)


def in6_addrtovendor(addr):
    """
    Extract the MAC address from a modified EUI-64 constructed IPv6
    address provided and use the IANA oui.txt file to get the vendor.
    The database used for the conversion is the one loaded by Scapy
    from a Wireshark installation if discovered in a well-known
    location. None is returned on error, "UNKNOWN" if the vendor is
    unknown.
    """
    mac = in6_addrtomac(addr)
    if mac is None or not conf.manufdb:
        return None

    res = conf.manufdb._get_manuf(mac)
    if len(res) == 17 and res.count(':') != 5:  # Mac address, i.e. unknown
        res = "UNKNOWN"

    return res


def in6_getLinkScopedMcastAddr(addr, grpid=None, scope=2):
    """
    Generate a Link-Scoped Multicast Address as described in RFC 4489.
    Returned value is in printable notation.

    'addr' parameter specifies the link-local address to use for generating
    Link-scoped multicast address IID.

    By default, the function returns a ::/96 prefix (aka last 32 bits of
    returned address are null). If a group id is provided through 'grpid'
    parameter, last 32 bits of the address are set to that value (accepted
    formats : b'\x12\x34\x56\x78' or '12345678' or 0x12345678 or 305419896).

    By default, generated address scope is Link-Local (2). That value can
    be modified by passing a specific 'scope' value as an argument of the
    function. RFC 4489 only authorizes scope values <= 2. Enforcement
    is performed by the function (None will be returned).

    If no link-local address can be used to generate the Link-Scoped IPv6
    Multicast address, or if another error occurs, None is returned.
    """
    if scope not in [0, 1, 2]:
        return None
    try:
        if not in6_islladdr(addr):
            return None
        addr = inet_pton(socket.AF_INET6, addr)
    except Exception:
        warning("in6_getLinkScopedMcastPrefix(): Invalid address provided")
        return None

    iid = addr[8:]

    if grpid is None:
        grpid = b'\x00\x00\x00\x00'
    else:
        if isinstance(grpid, (bytes, str)):
            if len(grpid) == 8:
                try:
                    grpid = int(grpid, 16) & 0xffffffff
                except Exception:
                    warning("in6_getLinkScopedMcastPrefix(): Invalid group id provided")  # noqa: E501
                    return None
            elif len(grpid) == 4:
                try:
                    grpid = struct.unpack("!I", grpid)[0]
                except Exception:
                    warning("in6_getLinkScopedMcastPrefix(): Invalid group id provided")  # noqa: E501
                    return None
        grpid = struct.pack("!I", grpid)

    flgscope = struct.pack("B", 0xff & ((0x3 << 4) | scope))
    plen = b'\xff'
    res = b'\x00'
    a = b'\xff' + flgscope + res + plen + iid + grpid

    return inet_ntop(socket.AF_INET6, a)


def in6_get6to4Prefix(addr):
    """
    Returns the /48 6to4 prefix associated with provided IPv4 address
    On error, None is returned. No check is performed on public/private
    status of the address
    """
    try:
        addr = inet_pton(socket.AF_INET, addr)
        addr = inet_ntop(socket.AF_INET6, b'\x20\x02' + addr + b'\x00' * 10)
    except Exception:
        return None
    return addr


def in6_6to4ExtractAddr(addr):
    """
    Extract IPv4 address embedded in 6to4 address. Passed address must be
    a 6to4 address. None is returned on error.
    """
    try:
        addr = inet_pton(socket.AF_INET6, addr)
    except Exception:
        return None
    if addr[:2] != b" \x02":
        return None
    return inet_ntop(socket.AF_INET, addr[2:6])


def in6_getLocalUniquePrefix():
    """
    Returns a pseudo-randomly generated Local Unique prefix. Function
    follows recommendation of Section 3.2.2 of RFC 4193 for prefix
    generation.
    """
    # Extracted from RFC 1305 (NTP) :
    # NTP timestamps are represented as a 64-bit unsigned fixed-point number,
    # in seconds relative to 0h on 1 January 1900. The integer part is in the
    # first 32 bits and the fraction part in the last 32 bits.

    # epoch = (1900, 1, 1, 0, 0, 0, 5, 1, 0)
    # x = time.time()
    # from time import gmtime, strftime, gmtime, mktime
    # delta = mktime(gmtime(0)) - mktime(self.epoch)
    # x = x-delta

    tod = time.time()  # time of day. Will bother with epoch later
    i = int(tod)
    j = int((tod - i) * (2**32))
    tod = struct.pack("!II", i, j)
    mac = RandMAC()
    # construct modified EUI-64 ID
    eui64 = inet_pton(socket.AF_INET6, '::' + in6_mactoifaceid(mac))[8:]
    import hashlib
    globalid = hashlib.sha1(tod + eui64).digest()[:5]
    return inet_ntop(socket.AF_INET6, b'\xfd' + globalid + b'\x00' * 10)


def in6_getRandomizedIfaceId(ifaceid, previous=None):
    """
    Implements the interface ID generation algorithm described in RFC 3041.
    The function takes the Modified EUI-64 interface identifier generated
    as described in RFC 4291 and an optional previous history value (the
    first element of the output of this function). If no previous interface
    identifier is provided, a random one is generated. The function returns
    a tuple containing the randomized interface identifier and the history
    value (for possible future use). Input and output values are provided in
    a "printable" format as depicted below.

    ex::
        >>> in6_getRandomizedIfaceId('20b:93ff:feeb:2d3')
        ('4c61:76ff:f46a:a5f3', 'd006:d540:db11:b092')
        >>> in6_getRandomizedIfaceId('20b:93ff:feeb:2d3',
                                     previous='d006:d540:db11:b092')
        ('fe97:46fe:9871:bd38', 'eeed:d79c:2e3f:62e')
    """

    s = b""
    if previous is None:
        d = b"".join(chb(x) for x in range(256))
        for _ in range(8):
            s += chb(random.choice(d))
        previous = s
    s = inet_pton(socket.AF_INET6, "::" + ifaceid)[8:] + previous
    import hashlib
    s = hashlib.md5(s).digest()
    s1, s2 = s[:8], s[8:]
    s1 = chb(orb(s1[0]) | 0x04) + s1[1:]
    s1 = inet_ntop(socket.AF_INET6, b"\xff" * 8 + s1)[20:]
    s2 = inet_ntop(socket.AF_INET6, b"\xff" * 8 + s2)[20:]
    return (s1, s2)


_rfc1924map = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E',  # noqa: E501
               'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',  # noqa: E501
               'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',  # noqa: E501
               'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',  # noqa: E501
               'y', 'z', '!', '#', '$', '%', '&', '(', ')', '*', '+', '-', ';', '<', '=',  # noqa: E501
               '>', '?', '@', '^', '_', '`', '{', '|', '}', '~']


def in6_ctop(addr):
    """
    Convert an IPv6 address in Compact Representation Notation
    (RFC 1924) to printable representation ;-)
    Returns None on error.
    """
    if len(addr) != 20 or not reduce(lambda x, y: x and y,
                                     [x in _rfc1924map for x in addr]):
        return None
    i = 0
    for c in addr:
        j = _rfc1924map.index(c)
        i = 85 * i + j
    res = []
    for j in range(4):
        res.append(struct.pack("!I", i % 2**32))
        i = i // (2**32)
    res.reverse()
    return inet_ntop(socket.AF_INET6, b"".join(res))


def in6_ptoc(addr):
    """
    Converts an IPv6 address in printable representation to RFC
    1924 Compact Representation ;-)
    Returns None on error.
    """
    try:
        d = struct.unpack("!IIII", inet_pton(socket.AF_INET6, addr))
    except Exception:
        return None
    res = 0
    m = [2**96, 2**64, 2**32, 1]
    for i in range(4):
        res += d[i] * m[i]
    rem = res
    res = []
    while rem:
        res.append(_rfc1924map[rem % 85])
        rem = rem // 85
    res.reverse()
    return "".join(res)


def in6_isaddr6to4(x):
    """
    Return True if provided address (in printable format) is a 6to4
    address (being in 2002::/16).
    """
    x = inet_pton(socket.AF_INET6, x)
    return x[:2] == b' \x02'


conf.teredoPrefix = "2001::"  # old one was 3ffe:831f (it is a /32)
conf.teredoServerPort = 3544


def in6_isaddrTeredo(x):
    """
    Return True if provided address is a Teredo, meaning it is under
    the /32 conf.teredoPrefix prefix value (by default, 2001::).
    Otherwise, False is returned. Address must be passed in printable
    format.
    """
    our = inet_pton(socket.AF_INET6, x)[0:4]
    teredoPrefix = inet_pton(socket.AF_INET6, conf.teredoPrefix)[0:4]
    return teredoPrefix == our


def teredoAddrExtractInfo(x):
    """
    Extract information from a Teredo address. Return value is
    a 4-tuple made of IPv4 address of Teredo server, flag value (int),
    mapped address (non obfuscated) and mapped port (non obfuscated).
    No specific checks are performed on passed address.
    """
    addr = inet_pton(socket.AF_INET6, x)
    server = inet_ntop(socket.AF_INET, addr[4:8])
    flag = struct.unpack("!H", addr[8:10])[0]
    mappedport = struct.unpack("!H", strxor(addr[10:12], b'\xff' * 2))[0]
    mappedaddr = inet_ntop(socket.AF_INET, strxor(addr[12:16], b'\xff' * 4))
    return server, flag, mappedaddr, mappedport


def in6_iseui64(x):
    """
    Return True if provided address has an interface identifier part
    created in modified EUI-64 format (meaning it matches ``*::*:*ff:fe*:*``).
    Otherwise, False is returned. Address must be passed in printable
    format.
    """
    eui64 = inet_pton(socket.AF_INET6, '::ff:fe00:0')
    x = in6_and(inet_pton(socket.AF_INET6, x), eui64)
    return x == eui64


def in6_isanycast(x):  # RFC 2526
    if in6_iseui64(x):
        s = '::fdff:ffff:ffff:ff80'
        packed_x = inet_pton(socket.AF_INET6, x)
        packed_s = inet_pton(socket.AF_INET6, s)
        x_and_s = in6_and(packed_x, packed_s)
        return x_and_s == packed_s
    else:
        # not EUI-64
        # |              n bits             |    121-n bits    |   7 bits   |
        # +---------------------------------+------------------+------------+
        # |           subnet prefix         | 1111111...111111 | anycast ID |
        # +---------------------------------+------------------+------------+
        #                                   |   interface identifier field  |
        warning('in6_isanycast(): TODO not EUI-64')
        return 0


def _in6_bitops(a1, a2, operator=0):
    a1 = struct.unpack('4I', a1)
    a2 = struct.unpack('4I', a2)
    fop = [lambda x, y: x | y,
           lambda x, y: x & y,
           lambda x, y: x ^ y
           ]
    ret = map(fop[operator % len(fop)], a1, a2)
    return b"".join(struct.pack('I', x) for x in ret)


def in6_or(a1, a2):
    """
    Provides a bit to bit OR of provided addresses. They must be
    passed in network format. Return value is also an IPv6 address
    in network format.
    """
    return _in6_bitops(a1, a2, 0)


def in6_and(a1, a2):
    """
    Provides a bit to bit AND of provided addresses. They must be
    passed in network format. Return value is also an IPv6 address
    in network format.
    """
    return _in6_bitops(a1, a2, 1)


def in6_xor(a1, a2):
    """
    Provides a bit to bit XOR of provided addresses. They must be
    passed in network format. Return value is also an IPv6 address
    in network format.
    """
    return _in6_bitops(a1, a2, 2)


def in6_cidr2mask(m):
    """
    Return the mask (bitstring) associated with provided length
    value. For instance if function is called on 48, return value is
    b'\xff\xff\xff\xff\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'.

    """
    if m > 128 or m < 0:
        raise Scapy_Exception("value provided to in6_cidr2mask outside [0, 128] domain (%d)" % m)  # noqa: E501

    t = []
    for i in range(0, 4):
        t.append(max(0, 2**32 - 2**(32 - min(32, m))))
        m -= 32

    return b"".join(struct.pack('!I', x) for x in t)


def in6_getnsma(a):
    """
    Return link-local solicited-node multicast address for given
    address. Passed address must be provided in network format.
    Returned value is also in network format.
    """

    r = in6_and(a, inet_pton(socket.AF_INET6, '::ff:ffff'))
    r = in6_or(inet_pton(socket.AF_INET6, 'ff02::1:ff00:0'), r)
    return r


def in6_getnsmac(a):  # return multicast Ethernet address associated with multicast v6 destination  # noqa: E501
    """
    Return the multicast mac address associated with provided
    IPv6 address. Passed address must be in network format.
    """

    a = struct.unpack('16B', a)[-4:]
    mac = '33:33:'
    mac += ':'.join("%.2x" % x for x in a)
    return mac


def in6_getha(prefix):
    """
    Return the anycast address associated with all home agents on a given
    subnet.
    """
    r = in6_and(inet_pton(socket.AF_INET6, prefix), in6_cidr2mask(64))
    r = in6_or(r, inet_pton(socket.AF_INET6, '::fdff:ffff:ffff:fffe'))
    return inet_ntop(socket.AF_INET6, r)


def in6_ptop(str):
    """
    Normalizes IPv6 addresses provided in printable format, returning the
    same address in printable format. (2001:0db8:0:0::1 -> 2001:db8::1)
    """
    return inet_ntop(socket.AF_INET6, inet_pton(socket.AF_INET6, str))


def in6_isincluded(addr, prefix, plen):
    """
    Returns True when 'addr' belongs to prefix/plen. False otherwise.
    """
    temp = inet_pton(socket.AF_INET6, addr)
    pref = in6_cidr2mask(plen)
    zero = inet_pton(socket.AF_INET6, prefix)
    return zero == in6_and(temp, pref)


def in6_isllsnmaddr(str):
    """
    Return True if provided address is a link-local solicited node
    multicast address, i.e. belongs to ff02::1:ff00:0/104. False is
    returned otherwise.
    """
    temp = in6_and(b"\xff" * 13 + b"\x00" * 3, inet_pton(socket.AF_INET6, str))
    temp2 = b'\xff\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\xff\x00\x00\x00'
    return temp == temp2


def in6_isdocaddr(str):
    """
    Returns True if provided address in printable format belongs to
    2001:db8::/32 address space reserved for documentation (as defined
    in RFC 3849).
    """
    return in6_isincluded(str, '2001:db8::', 32)


def in6_islladdr(str):
    """
    Returns True if provided address in printable format belongs to
    _allocated_ link-local unicast address space (fe80::/10)
    """
    return in6_isincluded(str, 'fe80::', 10)


def in6_issladdr(str):
    """
    Returns True if provided address in printable format belongs to
    _allocated_ site-local address space (fec0::/10). This prefix has
    been deprecated, address being now reserved by IANA. Function
    will remain for historic reasons.
    """
    return in6_isincluded(str, 'fec0::', 10)


def in6_isuladdr(str):
    """
    Returns True if provided address in printable format belongs to
    Unique local address space (fc00::/7).
    """
    return in6_isincluded(str, 'fc00::', 7)

# TODO : we should see the status of Unique Local addresses against
#        global address space.
#        Up-to-date information is available through RFC 3587.
#        We should review function behavior based on its content.


def in6_isgladdr(str):
    """
    Returns True if provided address in printable format belongs to
    _allocated_ global address space (2000::/3). Please note that,
    Unique Local addresses (FC00::/7) are not part of global address
    space, and won't match.
    """
    return in6_isincluded(str, '2000::', 3)


def in6_ismaddr(str):
    """
    Returns True if provided address in printable format belongs to
    allocated Multicast address space (ff00::/8).
    """
    return in6_isincluded(str, 'ff00::', 8)


def in6_ismnladdr(str):
    """
    Returns True if address belongs to node-local multicast address
    space (ff01::/16) as defined in RFC
    """
    return in6_isincluded(str, 'ff01::', 16)


def in6_ismgladdr(str):
    """
    Returns True if address belongs to global multicast address
    space (ff0e::/16).
    """
    return in6_isincluded(str, 'ff0e::', 16)


def in6_ismlladdr(str):
    """
    Returns True if address belongs to link-local multicast address
    space (ff02::/16)
    """
    return in6_isincluded(str, 'ff02::', 16)


def in6_ismsladdr(str):
    """
    Returns True if address belongs to site-local multicast address
    space (ff05::/16). Site local address space has been deprecated.
    Function remains for historic reasons.
    """
    return in6_isincluded(str, 'ff05::', 16)


def in6_isaddrllallnodes(str):
    """
    Returns True if address is the link-local all-nodes multicast
    address (ff02::1).
    """
    return (inet_pton(socket.AF_INET6, "ff02::1") ==
            inet_pton(socket.AF_INET6, str))


def in6_isaddrllallservers(str):
    """
    Returns True if address is the link-local all-servers multicast
    address (ff02::2).
    """
    return (inet_pton(socket.AF_INET6, "ff02::2") ==
            inet_pton(socket.AF_INET6, str))


def in6_getscope(addr):
    """
    Returns the scope of the address.
    """
    if in6_isgladdr(addr) or in6_isuladdr(addr):
        scope = IPV6_ADDR_GLOBAL
    elif in6_islladdr(addr):
        scope = IPV6_ADDR_LINKLOCAL
    elif in6_issladdr(addr):
        scope = IPV6_ADDR_SITELOCAL
    elif in6_ismaddr(addr):
        if in6_ismgladdr(addr):
            scope = IPV6_ADDR_GLOBAL
        elif in6_ismlladdr(addr):
            scope = IPV6_ADDR_LINKLOCAL
        elif in6_ismsladdr(addr):
            scope = IPV6_ADDR_SITELOCAL
        elif in6_ismnladdr(addr):
            scope = IPV6_ADDR_LOOPBACK
        else:
            scope = -1
    elif addr == '::1':
        scope = IPV6_ADDR_LOOPBACK
    else:
        scope = -1
    return scope


def in6_get_common_plen(a, b):
    """
    Return common prefix length of IPv6 addresses a and b.
    """
    def matching_bits(byte1, byte2):
        for i in range(8):
            cur_mask = 0x80 >> i
            if (byte1 & cur_mask) != (byte2 & cur_mask):
                return i
        return 8

    tmpA = inet_pton(socket.AF_INET6, a)
    tmpB = inet_pton(socket.AF_INET6, b)
    for i in range(16):
        mbits = matching_bits(orb(tmpA[i]), orb(tmpB[i]))
        if mbits != 8:
            return 8 * i + mbits
    return 128


def in6_isvalid(address):
    """Return True if 'address' is a valid IPv6 address string, False
       otherwise."""

    try:
        socket.inet_pton(socket.AF_INET6, address)
        return True
    except Exception:
        return False


class Net6(Gen):  # syntax ex. fec0::/126
    """Generate a list of IPv6s from a network address or a name"""
    name = "ipv6"
    ip_regex = re.compile(r"^([a-fA-F0-9:]+)(/[1]?[0-3]?[0-9])?$")

    def __init__(self, net):
        self.repr = net

        tmp = net.split('/') + ["128"]
        if not self.ip_regex.match(net):
            tmp[0] = socket.getaddrinfo(tmp[0], None, socket.AF_INET6)[0][-1][0]  # noqa: E501

        netmask = int(tmp[1])
        self.net = inet_pton(socket.AF_INET6, tmp[0])
        self.mask = in6_cidr2mask(netmask)
        self.plen = netmask

    def _parse(self):
        def parse_digit(value, netmask):
            netmask = min(8, max(netmask, 0))
            value = int(value)
            return (value & (0xff << netmask),
                    (value | (0xff >> (8 - netmask))) + 1)

        self.parsed = [
            parse_digit(x, y) for x, y in zip(
                struct.unpack("16B", in6_and(self.net, self.mask)),
                (x - self.plen for x in range(8, 129, 8)),
            )
        ]

    def __iter__(self):
        self._parse()

        def rec(n, li):
            sep = ':' if n and n % 2 == 0 else ''
            if n == 16:
                return li
            return rec(n + 1, [y + sep + '%.2x' % i
                               # faster than '%s%s%.2x' % (y, sep, i)
                               for i in range(*self.parsed[n])
                               for y in li])

        return (in6_ptop(addr) for addr in iter(rec(0, [''])))

    def __iterlen__(self):
        self._parse()
        return reduce(operator.mul, ((y - x) for (x, y) in self.parsed), 1)

    def __str__(self):
        try:
            return next(self.__iter__())
        except (StopIteration, RuntimeError):
            return None

    def __eq__(self, other):
        return str(other) == str(self)

    def __ne__(self, other):
        return not self == other

    __hash__ = None

    def __repr__(self):
        return "Net6(%r)" % self.repr



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information

"""
/*
 * Copyright (c) 1982, 1986, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)if_ether.h	8.1 (Berkeley) 6/10/93
 */
"""

# This file contains data automatically generated using
# scapy/tools/generate_ethertypes.py
# based on OpenBSD public source.

DATA = b"""
#
# Ethernet frame types
#       This file describes some of the various Ethernet
#       protocol types that are used on Ethernet networks.
#
# This list could be found on:
#         http://www.iana.org/assignments/ethernet-numbers
#         http://www.iana.org/assignments/ieee-802-numbers
#
# <name>    <hexnumber> <alias1>...<alias35> #Comment
#
8023            0004                         # IEEE 802.3 packet
PUP             0200                         # Xerox PUP protocol - see 0A00
PUPAT           0200                         # PUP Address Translation - see 0A01
NS              0600                         # XNS
NSAT            0601                         # XNS Address Translation (3Mb only)
DLOG1           0660                         # DLOG (?)
DLOG2           0661                         # DLOG (?)
IPv4            0800                         # IP protocol
X75             0801                         # X.75 Internet
NBS             0802                         # NBS Internet
ECMA            0803                         # ECMA Internet
CHAOS           0804                         # CHAOSnet
X25             0805                         # X.25 Level 3
ARP             0806                         # Address resolution protocol
FRARP           0808                         # Frame Relay ARP (RFC1701)
VINES           0BAD                         # Banyan VINES
TRAIL           1000                         # Trailer packet
DCA             1234                         # DCA - Multicast
VALID           1600                         # VALID system protocol
RCL             1995                         # Datapoint Corporation (RCL lan protocol)
NBPCC           3C04                         # 3Com NBP Connect complete not registered
NBPDG           3C07                         # 3Com NBP Datagram (like XNS IDP) not registered
PCS             4242                         # PCS Basic Block Protocol
IMLBL           4C42                         # Information Modes Little Big LAN
MOPDL           6001                         # DEC MOP dump/load
MOPRC           6002                         # DEC MOP remote console
LAT             6004                         # DEC LAT
SCA             6007                         # DEC LAVC, SCA
AMBER           6008                         # DEC AMBER
RAWFR           6559                         # Raw Frame Relay (RFC1701)
UBDL            7000                         # Ungermann-Bass download
UBNIU           7001                         # Ungermann-Bass NIUs
UBNMC           7003                         # Ungermann-Bass ??? (NMC to/from UB Bridge)
UBBST           7005                         # Ungermann-Bass Bridge Spanning Tree
OS9             7007                         # OS/9 Microware
RACAL           7030                         # Racal-Interlan
HP              8005                         # HP Probe
TIGAN           802F                         # Tigan, Inc.
DECAM           8048                         # DEC Availability Manager for Distributed Systems DECamds (but someone at DEC says not)
VEXP            805B                         # Stanford V Kernel exp.
VPROD           805C                         # Stanford V Kernel prod.
ES              805D                         # Evans & Sutherland
VEECO           8067                         # Veeco Integrated Auto.
ATT             8069                         # AT&T
MATRA           807A                         # Matra
DDE             807B                         # Dansk Data Elektronik
MERIT           807C                         # Merit Internodal (or Univ of Michigan?)
ATALK           809B                         # AppleTalk
PACER           80C6                         # Pacer Software
SNA             80D5                         # IBM SNA Services over Ethernet
RETIX           80F2                         # Retix
AARP            80F3                         # AppleTalk AARP
VLAN            8100                         # IEEE 802.1Q VLAN tagging (XXX conflicts)
BOFL            8102                         # Wellfleet; BOFL (Breath OF Life) pkts [every 5-10 secs.]
HAYES           8130                         # Hayes Microcomputers (XXX which?)
VGLAB           8131                         # VG Laboratory Systems
IPX             8137                         # Novell (old) NetWare IPX (ECONFIG E option)
MUMPS           813F                         # M/MUMPS data sharing
FLIP            8146                         # Vrije Universiteit (NL) FLIP (Fast Local Internet Protocol)
NCD             8149                         # Network Computing Devices
ALPHA           814A                         # Alpha Micro
SNMP            814C                         # SNMP over Ethernet (see RFC1089)
XTP             817D                         # Protocol Engines XTP
SGITW           817E                         # SGI/Time Warner prop.
STP             8181                         # Scheduled Transfer STP, HIPPI-ST
IPv6            86DD                         # IP protocol version 6
RDP             8739                         # Control Technology Inc. RDP Without IP
MICP            873A                         # Control Technology Inc. Mcast Industrial Ctrl Proto.
IPAS            876C                         # IP Autonomous Systems (RFC1701)
SLOW            8809                         # 803.3ad slow protocols (LACP/Marker)
PPP             880B                         # PPP (obsolete by PPPOE)
MPLS            8847                         # MPLS Unicast
AXIS            8856                         # Axis Communications AB proprietary bootstrap/config
PPPOE           8864                         # PPP Over Ethernet Session Stage
PAE             888E                         # 802.1X Port Access Entity
AOE             88A2                         # ATA over Ethernet
QINQ            88A8                         # 802.1ad VLAN stacking
LLDP            88CC                         # Link Layer Discovery Protocol
PBB             88E7                         # 802.1Q Provider Backbone Bridging
XNSSM           9001                         # 3Com (Formerly Bridge Communications), XNS Systems Management
TCPSM           9002                         # 3Com (Formerly Bridge Communications), TCP/IP Systems Management
DEBNI           AAAA                         # DECNET? Used by VAX 6220 DEBNI
SONIX           FAF5                         # Sonix Arpeggio
VITAL           FF00                         # BBN VITAL-LanBridge cache wakeups
MAX             FFFF                         # Maximum valid ethernet type, reserved
"""



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

"""
Functions common to different architectures
"""

import ctypes
import socket
import struct
import sys
import time
from pocsuite3.thirdparty.scapy.consts import WINDOWS
from pocsuite3.thirdparty.scapy.config import conf
from pocsuite3.thirdparty.scapy.data import MTU, ARPHRD_TO_DLT
from pocsuite3.thirdparty.scapy.error import Scapy_Exception

if not WINDOWS:
    from fcntl import ioctl

# UTILS


def get_if(iff, cmd):
    """Ease SIOCGIF* ioctl calls"""

    sck = socket.socket()
    try:
        return ioctl(sck, cmd, struct.pack("16s16x", iff.encode("utf8")))
    finally:
        sck.close()


def get_if_raw_hwaddr(iff):
    """Get the raw MAC address of a local interface.

    This function uses SIOCGIFHWADDR calls, therefore only works
    on some distros.

    :param iff: the network interface name as a string
    :returns: the corresponding raw MAC address
    """
    from pocsuite3.thirdparty.scapy.solaris import SIOCGIFHWADDR
    return struct.unpack("16xh6s8x", get_if(iff, SIOCGIFHWADDR))

# SOCKET UTILS


def _select_nonblock(sockets, remain=None):
    """This function is called during sendrecv() routine to select
    the available sockets.
    """
    # pcap sockets aren't selectable, so we return all of them
    # and ask the selecting functions to use nonblock_recv instead of recv
    def _sleep_nonblock_recv(self):
        res = self.nonblock_recv()
        if res is None:
            time.sleep(conf.recv_poll_rate)
        return res
    # we enforce remain=None: don't wait.
    return sockets, _sleep_nonblock_recv

# BPF HANDLERS


def compile_filter(filter_exp, iface=None, linktype=None,
                   promisc=False):
    """Asks libpcap to parse the filter, then build the matching
    BPF bytecode.

    :param iface: if provided, use the interface to compile
    :param linktype: if provided, use the linktype to compile
    """
    try:
        from pocsuite3.thirdparty.scapy.winpcapy import (
            PCAP_ERRBUF_SIZE,
            pcap_open_live,
            pcap_compile,
            pcap_compile_nopcap,
            pcap_close
        )
        from pocsuite3.thirdparty.scapy.structures import bpf_program
    except OSError:
        raise ImportError(
            "libpcap is not available. Cannot compile filter !"
        )
    from ctypes import create_string_buffer
    bpf = bpf_program()
    bpf_filter = create_string_buffer(filter_exp.encode("utf8"))
    if not linktype:
        # Try to guess linktype to avoid root
        if not iface:
            if not conf.iface:
                raise Scapy_Exception(
                    "Please provide an interface or linktype!"
                )
            if WINDOWS:
                iface = conf.iface.pcap_name
            else:
                iface = conf.iface
        # Try to guess linktype to avoid requiring root
        try:
            arphd = get_if_raw_hwaddr(iface)[0]
            linktype = ARPHRD_TO_DLT.get(arphd)
        except Exception:
            # Failed to use linktype: use the interface
            pass
    if linktype is not None:
        ret = pcap_compile_nopcap(
            MTU, linktype, ctypes.byref(bpf), bpf_filter, 0, -1
        )
    elif iface:
        err = create_string_buffer(PCAP_ERRBUF_SIZE)
        iface = create_string_buffer(iface.encode("utf8"))
        pcap = pcap_open_live(
            iface, MTU, promisc, 0, err
        )
        error = bytes(bytearray(err)).strip(b"\x00")
        if error:
            raise OSError(error)
        ret = pcap_compile(
            pcap, ctypes.byref(bpf), bpf_filter, 0, -1
        )
        pcap_close(pcap)
    # if ret == -1:
    #     raise Scapy_Exception(
    #         "Failed to compile filter expression %s (%s)" % (filter_exp, ret)
    #     )
    if conf.use_pypy and sys.pypy_version_info <= (7, 3, 0):
        # PyPy < 7.3.0 has a broken behavior
        # https://bitbucket.org/pypy/pypy/issues/3114
        return struct.pack(
            'HL',
            bpf.bf_len, ctypes.addressof(bpf.bf_insns.contents)
        )
    return bpf



# This file is part of Scapy
# See http://www.secdev.org/projects/scapy for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>
# This program is published under a GPLv2 license

"""
Common customizations for all Unix-like operating systems other than Linux
"""

import os
import socket

from pocsuite3.thirdparty.scapy import config
from pocsuite3.thirdparty.scapy import utils
from pocsuite3.thirdparty.scapy.arch_init import get_if_addr
from pocsuite3.thirdparty.scapy.config import conf
from pocsuite3.thirdparty.scapy.consts import FREEBSD, NETBSD, OPENBSD, SOLARIS
from pocsuite3.thirdparty.scapy.error import warning, log_interactive
from pocsuite3.thirdparty.scapy.pton_ntop import inet_pton
from pocsuite3.thirdparty.scapy.utils6 import in6_getscope, construct_source_candidate_set
from pocsuite3.thirdparty.scapy.utils6 import in6_isvalid, in6_ismlladdr, in6_ismnladdr


##################
#  Routes stuff  #
##################

def _guess_iface_name(netif):
    """
    We attempt to guess the name of interfaces that are truncated from the
    output of ifconfig -l.
    If there is only one possible candidate matching the interface name then we
    return it.
    If there are none or more, then we return None.
    """
    with os.popen('%s -l' % conf.prog.ifconfig) as fdesc:
        ifaces = fdesc.readline().strip().split(' ')
    matches = [iface for iface in ifaces if iface.startswith(netif)]
    if len(matches) == 1:
        return matches[0]
    return None


def read_routes():
    """Return a list of IPv4 routes than can be used by Scapy.

    This function parses netstat.
    """
    if SOLARIS:
        f = os.popen("netstat -rvn -f inet")
    elif FREEBSD:
        f = os.popen("netstat -rnW -f inet")  # -W to show long interface names
    else:
        f = os.popen("netstat -rn -f inet")
    ok = 0
    mtu_present = False
    prio_present = False
    refs_present = False
    use_present = False
    routes = []
    pending_if = []
    for line in f.readlines():
        if not line:
            break
        line = line.strip().lower()
        if line.find("----") >= 0:  # a separation line
            continue
        if not ok:
            if line.find("destination") >= 0:
                ok = 1
                mtu_present = "mtu" in line
                prio_present = "prio" in line
                refs_present = "ref" in line  # There is no s on Solaris
                use_present = "use" in line or "nhop" in line
            continue
        if not line:
            break
        rt = line.split()
        if SOLARIS:
            dest, netmask, gw, netif = rt[:4]
            flg = rt[4 + mtu_present + refs_present]
        else:
            dest, gw, flg = rt[:3]
            locked = OPENBSD and rt[6] == "l"
            offset = mtu_present + prio_present + refs_present + locked
            offset += use_present
            netif = rt[3 + offset]
        if flg.find("lc") >= 0:
            continue
        elif dest == "default":
            dest = 0
            netmask = 0
        elif SOLARIS:
            dest = utils.atol(dest)
            netmask = utils.atol(netmask)
        else:
            if "/" in dest:
                dest, netmask = dest.split("/")
                netmask = utils.itom(int(netmask))
            else:
                netmask = utils.itom((dest.count(".") + 1) * 8)
            dest += ".0" * (3 - dest.count("."))
            dest = utils.atol(dest)
        # XXX: TODO: add metrics for unix.py (use -e option on netstat)
        metric = 1
        if "g" not in flg:
            gw = '0.0.0.0'
        if netif is not None:
            try:
                ifaddr = get_if_addr(netif)
                routes.append((dest, netmask, gw, netif, ifaddr, metric))
            except OSError as exc:
                if exc.message == 'Device not configured':
                    # This means the interface name is probably truncated by
                    # netstat -nr. We attempt to guess it's name and if not we
                    # ignore it.
                    guessed_netif = _guess_iface_name(netif)
                    if guessed_netif is not None:
                        ifaddr = get_if_addr(guessed_netif)
                        routes.append((dest, netmask, gw, guessed_netif, ifaddr, metric))  # noqa: E501
                    else:
                        warning("Could not guess partial interface name: %s", netif)  # noqa: E501
                else:
                    raise
        else:
            pending_if.append((dest, netmask, gw))
    f.close()

    # On Solaris, netstat does not provide output interfaces for some routes
    # We need to parse completely the routing table to route their gw and
    # know their output interface
    for dest, netmask, gw in pending_if:
        gw_l = utils.atol(gw)
        max_rtmask, gw_if, gw_if_addr, = 0, None, None
        for rtdst, rtmask, _, rtif, rtaddr in routes[:]:
            if gw_l & rtmask == rtdst:
                if rtmask >= max_rtmask:
                    max_rtmask = rtmask
                    gw_if = rtif
                    gw_if_addr = rtaddr
        # XXX: TODO add metrics
        metric = 1
        if gw_if:
            routes.append((dest, netmask, gw, gw_if, gw_if_addr, metric))
        else:
            warning("Did not find output interface to reach gateway %s", gw)

    return routes

############
#   IPv6   #
############


def _in6_getifaddr(ifname):
    """
    Returns a list of IPv6 addresses configured on the interface ifname.
    """

    # Get the output of ifconfig
    try:
        f = os.popen("%s %s" % (conf.prog.ifconfig, ifname))
    except OSError:
        log_interactive.warning("Failed to execute ifconfig.")
        return []

    # Iterate over lines and extract IPv6 addresses
    ret = []
    for line in f:
        if "inet6" in line:
            addr = line.rstrip().split(None, 2)[1]  # The second element is the IPv6 address  # noqa: E501
        else:
            continue
        if '%' in line:  # Remove the interface identifier if present
            addr = addr.split("%", 1)[0]

        # Check if it is a valid IPv6 address
        try:
            inet_pton(socket.AF_INET6, addr)
        except (socket.error, ValueError):
            continue

        # Get the scope and keep the address
        scope = in6_getscope(addr)
        ret.append((addr, scope, ifname))

    f.close()
    return ret


def in6_getifaddr():
    """
    Returns a list of 3-tuples of the form (addr, scope, iface) where
    'addr' is the address of scope 'scope' associated to the interface
    'iface'.

    This is the list of all addresses of all interfaces available on
    the system.
    """

    # List all network interfaces
    if OPENBSD or SOLARIS:
        if SOLARIS:
            cmd = "%s -a6"
        else:
            cmd = "%s"
        try:
            f = os.popen(cmd % conf.prog.ifconfig)
        except OSError:
            log_interactive.warning("Failed to execute ifconfig.")
            return []

        # Get the list of network interfaces
        splitted_line = []
        for line in f:
            if "flags" in line:
                iface = line.split()[0].rstrip(':')
                splitted_line.append(iface)

    else:  # FreeBSD, NetBSD or Darwin
        try:
            f = os.popen("%s -l" % conf.prog.ifconfig)
        except OSError:
            log_interactive.warning("Failed to execute ifconfig.")
            return []

        # Get the list of network interfaces
        splitted_line = f.readline().rstrip().split()

    ret = []
    for i in splitted_line:
        ret += _in6_getifaddr(i)
    f.close()
    return ret


def read_routes6():
    """Return a list of IPv6 routes than can be used by Scapy.

    This function parses netstat.
    """

    # Call netstat to retrieve IPv6 routes
    fd_netstat = os.popen("netstat -rn -f inet6")

    # List interfaces IPv6 addresses
    lifaddr = in6_getifaddr()
    if not lifaddr:
        fd_netstat.close()
        return []

    # Routes header information
    got_header = False
    mtu_present = False
    prio_present = False

    # Parse the routes
    routes = []
    for line in fd_netstat.readlines():

        # Parse the routes header and try to identify extra columns
        if not got_header:
            if "Destination" == line[:11]:
                got_header = True
                mtu_present = "Mtu" in line
                prio_present = "Prio" in line
            continue

        # Parse a route entry according to the operating system
        splitted_line = line.split()
        if OPENBSD or NETBSD:
            index = 5 + mtu_present + prio_present
            if len(splitted_line) < index:
                warning("Not enough columns in route entry !")
                continue
            destination, next_hop, flags = splitted_line[:3]
            dev = splitted_line[index]
        else:
            # FREEBSD or DARWIN
            if len(splitted_line) < 4:
                warning("Not enough columns in route entry !")
                continue
            destination, next_hop, flags, dev = splitted_line[:4]

        # XXX: TODO: add metrics for unix.py (use -e option on netstat)
        metric = 1

        # Check flags
        if "U" not in flags:  # usable route
            continue
        if "R" in flags:  # Host or net unreachable
            continue
        if "m" in flags:  # multicast address
            # Note: multicast routing is handled in Route6.route()
            continue

        # Replace link with the default route in next_hop
        if "link" in next_hop:
            next_hop = "::"

        # Default prefix length
        destination_plen = 128

        # Extract network interface from the zone id
        if '%' in destination:
            destination, dev = destination.split('%')
            if '/' in dev:
                # Example: fe80::%lo0/64 ; dev = "lo0/64"
                dev, destination_plen = dev.split('/')
        if '%' in next_hop:
            next_hop, dev = next_hop.split('%')

        # Ensure that the next hop is a valid IPv6 address
        if not in6_isvalid(next_hop):
            # Note: the 'Gateway' column might contain a MAC address
            next_hop = "::"

        # Modify parsed routing entries
        # Note: these rules are OS specific and may evolve over time
        if destination == "default":
            destination, destination_plen = "::", 0
        elif '/' in destination:
            # Example: fe80::/10
            destination, destination_plen = destination.split('/')
        if '/' in dev:
            # Example: ff02::%lo0/32 ; dev = "lo0/32"
            dev, destination_plen = dev.split('/')

        # Check route entries parameters consistency
        if not in6_isvalid(destination):
            warning("Invalid destination IPv6 address in route entry !")
            continue
        try:
            destination_plen = int(destination_plen)
        except Exception:
            warning("Invalid IPv6 prefix length in route entry !")
            continue
        if in6_ismlladdr(destination) or in6_ismnladdr(destination):
            # Note: multicast routing is handled in Route6.route()
            continue

        if conf.loopback_name in dev:
            # Handle ::1 separately
            cset = ["::1"]
            next_hop = "::"
        else:
            # Get possible IPv6 source addresses
            devaddrs = (x for x in lifaddr if x[2] == dev)
            cset = construct_source_candidate_set(destination, destination_plen, devaddrs)  # noqa: E501

        if len(cset):
            routes.append((destination, destination_plen, next_hop, dev, cset, metric))  # noqa: E501

    fd_netstat.close()
    return routes



from pocsuite3.lib.core.common import data_to_stdout
from pocsuite3.thirdparty.prettytable.prettytable import PrettyTable


class StatisticsComparison(object):

    def __init__(self):
        self.data = {}
        self.dork = {}

    def add_dork(self, source, dork):
        self.dork[source] = dork

    def add_ip(self, ip, source, honeypot=False):
        if ip not in self.data:
            self.data[ip] = {
                "source": [],
                "honeypot": honeypot,
                "success": False
            }
        self.data[ip]["source"].append(source)

    def getinfo(self, ip) -> tuple:
        if ip not in self.data:
            return "Other", "Unknown"
        sources = self.data[ip]["source"]
        return ','.join(sources), str(self.data[ip]["honeypot"])

    def change_success(self, ip, success=False):
        if ip in self.data:
            self.data[ip]["success"] = success

    def _statistics(self) -> dict:
        static_data = {}
        for ip, item in self.data.items():
            engines = item["source"]
            for engine in engines:
                if engine not in static_data:
                    static_data[engine] = {
                        "total": 0,
                        "success": 0,
                        "repetition": 0
                    }
                static_data[engine]["total"] += 1
                if item["success"]:
                    static_data[engine]["success"] += 1
                if len(engines) > 1:
                    static_data[engine]["repetition"] += 1
        return static_data

    def output(self):
        results_table = PrettyTable(["Search-engine", "Dork", "Total-data", "Success-rate", "Repetition-rate"])
        results_table.align["Search-engine"] = "c"
        results_table.padding_width = 1
        results = []
        for engine, item in self._statistics().items():
            dork = ""
            if engine in self.dork:
                dork = self.dork[engine]
            _result = [
                engine,
                dork,
                item["total"],
                "{0:.1f}%".format(item["success"] / item["total"] * 100),
                "{0:.1f}%".format(item["repetition"] / item["total"] * 100)
            ]
            results.append(_result)

        for row in results:
            results_table.add_row(row)

        data_to_stdout('\n{0}\n'.format(results_table.get_string()))
